import torch
import torch.nn as nn
import torch.nn.functional as F
from .gma import (
    Aggregate,
    SpatioTemporalAggregate,
    TemporalAggregate,
    TMMAggregate,
)

# from core.models.sk_decoder import vis_featmap

from torch import nn
from einops import rearrange

# from models.gaflow_modules.modules import GGAM


class PCBlock4_Deep_nopool_res(nn.Module):
    def __init__(self, C_in, C_out, k_conv):
        super().__init__()
        self.conv_list = nn.ModuleList(
            [
                nn.Conv2d(
                    C_in, C_in, kernel, stride=1, padding=kernel // 2, groups=C_in
                )
                for kernel in k_conv
            ]
        )

        self.ffn1 = nn.Sequential(
            nn.Conv2d(C_in, int(1.5 * C_in), 1, padding=0),
            nn.GELU(),
            nn.Conv2d(int(1.5 * C_in), C_in, 1, padding=0),
        )
        self.pw = nn.Conv2d(C_in, C_in, 1, padding=0)
        self.ffn2 = nn.Sequential(
            nn.Conv2d(C_in, int(1.5 * C_in), 1, padding=0),
            nn.GELU(),
            nn.Conv2d(int(1.5 * C_in), C_out, 1, padding=0),
        )

    def forward(self, x):
        x = F.gelu(x + self.ffn1(x))
        for conv in self.conv_list:
            x = F.gelu(x + conv(x))
        x = F.gelu(x + self.pw(x))
        x = self.ffn2(x)
        return x


class SKBlock_Temporal(nn.Module):
    def __init__(self, C_in, C_out, k_conv):
        super().__init__()
        self.conv_list = nn.ModuleList(
            [
                nn.Conv1d(
                    C_in, C_in, kernel, stride=1, padding=kernel // 2, groups=C_in
                )
                for kernel in k_conv
            ]
        )

        self.ffn1 = nn.Sequential(
            nn.Conv1d(C_in, int(1.5 * C_in), 1, padding=0),
            nn.GELU(),
            nn.Conv1d(int(1.5 * C_in), C_in, 1, padding=0),
        )
        self.pw = nn.Conv1d(C_in, C_in, 1, padding=0)
        self.ffn2 = nn.Sequential(
            nn.Conv1d(C_in, int(1.5 * C_in), 1, padding=0),
            nn.GELU(),
            nn.Conv1d(int(1.5 * C_in), C_out, 1, padding=0),
        )

    def forward(self, x, T=2):
        _, C, H, W = x.shape
        x = rearrange(x, "(b t) c h w -> (b h w) c t", t=T)
        # print(x.shape)
        x = F.gelu(x + self.ffn1(x))
        for conv in self.conv_list:
            x = F.gelu(x + conv(x))
        x = F.gelu(x + self.pw(x))
        x = self.ffn2(x)
        x = rearrange(x, "(b h w) c t -> (b t) c h w", h=H, w=W)

        return x


# class SKUpdate_GGAM(nn.Module):
#     def __init__(self, args):
#         super().__init__()
#         self.args = args
#         self.encoder = SKMotionEncoder6_Deep_nopool_res(args)
#         ratio = 16 if args.Encoder == 'UMT' else 8

#         self.gma = args.use_gma
#         embed_dim = args.decoder_dim // 2

#         if self.gma:
#             self.aggregator = GGAM(self.args, 128)
#             self.gru = PCBlock4_Deep_nopool_res(embed_dim*4, embed_dim, k_conv=args.PCUpdater_conv)
#         else:
#             self.gru = PCBlock4_Deep_nopool_res(embed_dim*4, embed_dim, k_conv=args.PCUpdater_conv)

#         self.flow_head = PCBlock4_Deep_nopool_res(embed_dim, 2, args.k_conv)

#         self.mask = nn.Sequential(
#             nn.Conv2d(embed_dim, embed_dim*2, 3, padding=1),
#             nn.ReLU(inplace=True),
#             nn.Conv2d(embed_dim*2, ratio*ratio*9, 1, padding=0))

#     def forward(self, net, inp, corr, flow, iters):
#         motion_features = self.encoder(flow, corr)

#         if self.gma:
#             motion_features_global = self.aggregator(inp, motion_features, iters)
#             inp_cat = torch.cat([inp, motion_features, motion_features_global], dim=1)
#         else:
#             inp_cat = torch.cat([inp, motion_features], dim=1)
#         # Attentional update
#         net = self.gru(torch.cat([net, inp_cat], dim=1))

#         delta_flow = self.flow_head(net)

#         # scale mask to balence gradients
#         mask = .25 * self.mask(net)
#         return net, mask, delta_flow


class SKBlock_Dilated(nn.Module):
    def __init__(self, C_in, C_out, k_conv, args=None):
        super().__init__()
        self.conv_list = nn.ModuleList(
            [
                nn.Conv2d(
                    C_in, C_in, kernel, stride=1, padding=kernel // 2, groups=C_in
                )
                for kernel in k_conv
            ]
        )

        self.dilated = nn.Conv2d(
            C_in,
            C_in,
            args.dilated_kernel,
            stride=1,
            padding=2 * (args.dilated_kernel - 1) // 2,
            groups=C_in,
            dilation=args.dilation_rate,
        )
        self.ffn1 = nn.Sequential(
            nn.Conv2d(C_in, int(1.5 * C_in), 1, padding=0),
            nn.GELU(),
            nn.Conv2d(int(1.5 * C_in), C_in, 1, padding=0),
        )
        self.pw = nn.Conv2d(C_in, C_in, 1, padding=0)
        self.ffn2 = nn.Sequential(
            nn.Conv2d(C_in, int(1.5 * C_in), 1, padding=0),
            nn.GELU(),
            nn.Conv2d(int(1.5 * C_in), C_out, 1, padding=0),
        )

    def forward(self, x):
        x = F.gelu(x + self.ffn1(x))
        for conv in self.conv_list:
            x = F.gelu(x + conv(x))
        x = F.gelu(x + self.pw(x))
        x = F.gelu(x + self.dilated(x))
        x = self.ffn2(x)
        return x


class SKBlock_Dilated_v2(nn.Module):
    def __init__(self, C_in, C_out, k_conv, args=None):
        super().__init__()
        self.conv_list = nn.ModuleList(
            [
                nn.Conv2d(
                    C_in, C_in, kernel, stride=1, padding=kernel // 2, groups=C_in
                )
                for kernel in k_conv
            ]
        )

        self.dilated = nn.Conv2d(
            C_in,
            C_in,
            args.dilated_kernel,
            stride=1,
            padding=2 * (args.dilated_kernel - 1) // 2,
            groups=C_in,
            dilation=args.dilation_rate,
        )
        self.ffn1 = nn.Sequential(
            nn.Conv2d(C_in, int(1.5 * C_in), 1, padding=0),
            nn.GELU(),
            nn.Conv2d(int(1.5 * C_in), C_in, 1, padding=0),
        )
        self.pw = nn.Conv2d(C_in, C_in, 1, padding=0)
        self.ffn2 = nn.Sequential(
            nn.Conv2d(C_in, int(1.5 * C_in), 1, padding=0),
            nn.GELU(),
            nn.Conv2d(int(1.5 * C_in), C_out, 1, padding=0),
        )

    def forward(self, x):
        dil_x = F.gelu(self.dilated(x))
        x = F.gelu(x + self.ffn1(x))
        for conv in self.conv_list:
            x = F.gelu(x + conv(x))
        x = F.gelu(x + self.pw(x))
        x = F.gelu(x + dil_x)
        x = self.ffn2(x)
        return x


class SKBlock_Dilated_v3(nn.Module):
    def __init__(self, C_in, C_out, k_conv, args=None):
        super().__init__()
        self.conv_list = nn.ModuleList(
            [
                nn.Conv2d(
                    C_in, C_in, kernel, stride=1, padding=kernel // 2, groups=C_in
                )
                for kernel in k_conv
            ]
        )

        self.dilated = nn.Conv2d(
            C_in,
            C_in,
            args.dilated_kernel,
            stride=1,
            padding=2 * (args.dilated_kernel - 1) // 2,
            groups=C_in,
            dilation=args.dilation_rate,
        )
        self.ffn1 = nn.Sequential(
            nn.Conv2d(C_in, int(1.5 * C_in), 1, padding=0),
            nn.GELU(),
            nn.Conv2d(int(1.5 * C_in), C_in, 1, padding=0),
        )
        self.pw = nn.Conv2d(C_in, C_in, 1, padding=0)
        self.ffn2 = nn.Sequential(
            nn.Conv2d(C_in, int(1.5 * C_in), 1, padding=0),
            nn.GELU(),
            nn.Conv2d(int(1.5 * C_in), C_out, 1, padding=0),
        )

    def forward(self, x):
        x = F.gelu(x + self.ffn1(x))
        dil_x = F.gelu(self.dilated(x))
        for conv in self.conv_list:
            x = F.gelu(x + conv(x))
        x = F.gelu(x + self.pw(x))
        x = F.gelu(x + dil_x)
        x = self.ffn2(x)
        return x


class SKBlock_Dilated_v4(nn.Module):
    def __init__(self, C_in, C_out, k_conv, args=None):
        super().__init__()
        self.conv_list = nn.ModuleList(
            [
                nn.Conv2d(
                    C_in, C_in, kernel, stride=1, padding=kernel // 2, groups=C_in
                )
                for kernel in k_conv
            ]
        )

        self.dilated1 = nn.Conv2d(
            C_in,
            C_in,
            9,
            stride=1,
            padding=2 * (9 - 1) // 2,
            groups=C_in,
            dilation=args.dilation_rate,
        )
        self.dilated2 = nn.Conv2d(
            C_in,
            C_in,
            5,
            stride=1,
            padding=2 * (5 - 1) // 2,
            groups=C_in,
            dilation=args.dilation_rate,
        )
        self.ffn1 = nn.Sequential(
            nn.Conv2d(C_in, int(1.5 * C_in), 1, padding=0),
            nn.GELU(),
            nn.Conv2d(int(1.5 * C_in), C_in, 1, padding=0),
        )
        self.pw = nn.Conv2d(C_in, C_in, 1, padding=0)
        self.ffn2 = nn.Sequential(
            nn.Conv2d(C_in, int(1.5 * C_in), 1, padding=0),
            nn.GELU(),
            nn.Conv2d(int(1.5 * C_in), C_out, 1, padding=0),
        )

    def forward(self, x):
        x = F.gelu(x + self.ffn1(x))
        dil_x1 = F.gelu(self.dilated1(x))
        dil_x2 = F.gelu(self.dilated2(x))
        for conv in self.conv_list:
            x = F.gelu(x + conv(x))
        x = F.gelu(x + self.pw(x))
        x = F.gelu(x + dil_x1 + dil_x2)
        x = self.ffn2(x)
        return x


class SKBlock_Dilated_multi(nn.Module):
    def __init__(self, C_in, C_out, k_conv, args=None):
        super().__init__()
        self.conv_list = nn.ModuleList(
            [
                nn.Conv2d(
                    C_in, C_in, kernel, stride=1, padding=kernel // 2, groups=C_in
                )
                for kernel in k_conv
            ]
        )

        self.perceptual_list = nn.ModuleList(
            [
                nn.Conv2d(
                    C_in, C_in, kernel, stride=1, padding=kernel // 2, groups=C_in
                )
                for kernel in args.perceptuals
            ]
        )

        self.dilated_list = nn.ModuleList(
            [
                nn.Conv2d(
                    C_in,
                    C_in,
                    kernel,
                    stride=1,
                    padding=2 * (kernel - 1) // 2,
                    groups=C_in,
                    dilation=args.dilation_rate,
                )
                for kernel in args.dilated_kernels
            ]
        )

        self.ffn1 = nn.Sequential(
            nn.Conv2d(C_in, int(1.5 * C_in), 1, padding=0),
            nn.GELU(),
            nn.Conv2d(int(1.5 * C_in), C_in, 1, padding=0),
        )
        self.pw = nn.Conv2d(C_in, C_in, 1, padding=0)
        self.ffn2 = nn.Sequential(
            nn.Conv2d(C_in, int(1.5 * C_in), 1, padding=0),
            nn.GELU(),
            nn.Conv2d(int(1.5 * C_in), C_out, 1, padding=0),
        )

    def forward(self, x):
        x = F.gelu(x + self.ffn1(x))
        res = x

        x = F.gelu(x + sum(conv(x) for conv in self.perceptual_list))
        x = F.gelu(x + sum(conv(x) for conv in self.conv_list))
        x = F.gelu(x + self.pw(x))

        x = F.gelu(x + sum(F.gelu(conv(res)) for conv in self.dilated_list))
        x = self.ffn2(x)
        return x


class SKBlock_Dilated_v5(nn.Module):
    def __init__(self, C_in, C_out, k_conv, args=None):
        super().__init__()
        self.conv_list = nn.ModuleList(
            [
                nn.Conv2d(
                    C_in, C_in, kernel, stride=1, padding=kernel // 2, groups=C_in
                )
                for kernel in k_conv
            ]
        )

        self.aux = nn.Conv2d(C_in, C_in, 1, stride=1, padding=0)
        self.dilated1 = nn.Conv2d(
            C_in,
            C_in,
            9,
            stride=1,
            padding=2 * (9 - 1) // 2,
            groups=C_in,
            dilation=args.dilation_rate,
        )
        self.dilated2 = nn.Conv2d(
            C_in,
            C_in,
            5,
            stride=1,
            padding=2 * (5 - 1) // 2,
            groups=C_in,
            dilation=args.dilation_rate,
        )
        self.ffn1 = nn.Sequential(
            nn.Conv2d(C_in, int(1.5 * C_in), 1, padding=0),
            nn.GELU(),
            nn.Conv2d(int(1.5 * C_in), C_in, 1, padding=0),
        )
        self.pw = nn.Conv2d(C_in, C_in, 1, padding=0)
        self.ffn2 = nn.Sequential(
            nn.Conv2d(C_in, int(1.5 * C_in), 1, padding=0),
            nn.GELU(),
            nn.Conv2d(int(1.5 * C_in), C_out, 1, padding=0),
        )

    def forward(self, x):
        x = F.gelu(x + self.ffn1(x))
        dil_x1 = F.gelu(self.dilated1(x))
        dil_x2 = F.gelu(self.dilated2(x))
        x3 = F.gelu(self.aux(x))
        for conv in self.conv_list:
            x = F.gelu(x + conv(x))
        x = F.gelu(x + self.pw(x))
        x = F.gelu(x + dil_x1 + dil_x2 + x3)
        x = self.ffn2(x)
        return x


class SKMotionEncoder6_Deep_nopool_res(nn.Module):
    def __init__(self, decoder_dim, corr_levels, corr_radius, k_conv):
        super().__init__()
        out_dim = decoder_dim // 2
        cor_planes = corr_levels * (2 * corr_radius + 1) ** 2

        self.convc1 = PCBlock4_Deep_nopool_res(cor_planes, 256, k_conv)
        self.convc2 = PCBlock4_Deep_nopool_res(256, 192, k_conv)

        self.convf1 = nn.Conv2d(2, 128, 1, 1, 0)
        self.convf2 = PCBlock4_Deep_nopool_res(128, 64, k_conv)

        self.conv = PCBlock4_Deep_nopool_res(64 + 192, out_dim - 2, k_conv)

    def forward(self, flow, corr, attention=None):
        cor = F.gelu(self.convc1(corr))
        cor = self.convc2(cor)

        flo = self.convf1(flow)
        flo = self.convf2(flo)

        cor_flo = torch.cat([cor, flo], dim=1)
        out = self.conv(cor_flo)

        return torch.cat([out, flow], dim=1)


class SKMotionEncoder_CorrTemporal(nn.Module):
    def __init__(self, args):
        super().__init__()
        self.args = args
        out_dim = args.decoder_dim // 2
        cor_planes = args.corr_levels * (2 * args.corr_radius + 1) ** 2

        self.convc1 = PCBlock4_Deep_nopool_res(
            cor_planes, 256 // (args.T - 1), args.k_conv
        )
        self.convc2 = PCBlock4_Deep_nopool_res(256, 192 * (args.T - 1), args.k_conv)

        self.convf1 = nn.Conv2d(2 * (args.T - 1), 128, 1, 1, 0)
        self.convf2 = PCBlock4_Deep_nopool_res(128, 64 * (args.T - 1), args.k_conv)

        self.conv = PCBlock4_Deep_nopool_res(64 + 192, out_dim - 2, args.k_conv)

    def forward(self, flow, corr, attention=None):
        cor = F.gelu(self.convc1(corr))
        _, C, H, W = cor.shape
        cor = rearrange(cor, "(B T) C H W -> B (T C) H W", T=self.args.T - 1)
        cor = self.convc2(cor)

        cor = rearrange(cor, "B (T C) H W -> (B T) C H W", T=self.args.T - 1)

        flo = self.convf1(
            rearrange(flow, "(B T) C H W -> B (T C) H W", T=self.args.T - 1)
        )
        flo = self.convf2(flo)
        flo = rearrange(flo, "B (T C) H W -> (B T) C H W", T=self.args.T - 1)

        cor_flo = torch.cat([cor, flo], dim=1)
        out = self.conv(cor_flo)

        return torch.cat([out, flow], dim=1)


class SKBlock_Dilated_v6(nn.Module):
    def __init__(self, C_in, C_out, k_conv, args=None):
        super().__init__()
        self.conv_list = nn.ModuleList(
            [
                nn.Conv2d(
                    C_in, C_in, kernel, stride=1, padding=kernel // 2, groups=C_in
                )
                for kernel in k_conv
            ]
        )

        self.dilated = nn.Conv2d(
            C_in,
            C_in,
            args.dilated_kernel,
            stride=1,
            padding=2 * (args.dilated_kernel - 1) // 2,
            groups=C_in,
            dilation=args.dilation_rate,
        )
        self.aux = nn.Conv2d(C_in, C_in, 1, padding=0)
        self.ffn1 = nn.Sequential(
            nn.Conv2d(C_in, int(1.5 * C_in), 1, padding=0),
            nn.GELU(),
            nn.Conv2d(int(1.5 * C_in), C_in, 1, padding=0),
        )
        self.pw = nn.Conv2d(C_in, C_in, 1, padding=0)
        self.ffn2 = nn.Sequential(
            nn.Conv2d(C_in, int(1.5 * C_in), 1, padding=0),
            nn.GELU(),
            nn.Conv2d(int(1.5 * C_in), C_out, 1, padding=0),
        )

    def forward(self, x):
        x = F.gelu(x + self.ffn1(x))
        dil_x = F.gelu(self.dilated(x))
        aux_x = F.gelu(self.aux(x))
        for conv in self.conv_list:
            x = F.gelu(x + conv(x))
        x = F.gelu(x + self.pw(x))
        x = F.gelu(x + dil_x + aux_x)
        x = self.ffn2(x)
        return x


class SKUpdateBlock6_Deep_nopoolres_AllDecoder(nn.Module):
    def __init__(self, args):
        super().__init__()
        self.args = args
        self.encoder = SKMotionEncoder6_Deep_nopool_res(args)
        ratio = 16 if args.Encoder == "UMT" else 8

        self.gma = args.use_gma
        embed_dim = args.decoder_dim // 2

        if self.gma:
            self.aggregator = Aggregate(
                args=self.args,
                dim=embed_dim,
                dim_head=embed_dim,
                heads=self.args.num_heads,
            )
            self.gru = PCBlock4_Deep_nopool_res(
                embed_dim * 4, embed_dim, k_conv=args.PCUpdater_conv
            )
        else:
            self.gru = PCBlock4_Deep_nopool_res(
                embed_dim * 4, embed_dim, k_conv=args.PCUpdater_conv
            )

        self.flow_head = PCBlock4_Deep_nopool_res(embed_dim, 2, args.k_conv)

        self.mask = nn.Sequential(
            nn.Conv2d(embed_dim, embed_dim * 2, 3, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(embed_dim * 2, ratio * ratio * 9, 1, padding=0),
        )

    def forward(self, net, inp, corr, flow, attention, ref_nets=None):
        motion_features = self.encoder(flow, corr, attention)

        if self.gma:
            motion_features_global = self.aggregator(attention, motion_features)
            inp_cat = torch.cat([inp, motion_features, motion_features_global], dim=1)
        else:
            inp_cat = torch.cat([inp, motion_features], dim=1)
        # Attentional update
        net = self.gru(torch.cat([net, inp_cat], dim=1))

        delta_flow = self.flow_head(net)

        # scale mask to balence gradients
        mask = 0.25 * self.mask(net)
        return net, mask, delta_flow


from timm.models.vision_transformer import Attention as timm_attn
from timm.models.layers import DropPath, Mlp


def zero_module(module):
    # Zero out the parameters of a module and return it.
    for p in module.parameters():
        p.detach().zero_()
    return module


class TransformerBlock(nn.Module):
    def __init__(self, dim, num_heads=1, mlp_ratio=2, drop_rate=0.0):
        super().__init__()
        self.norm1 = nn.LayerNorm(dim)
        self.norm2 = nn.LayerNorm(dim)
        self.drop_path1 = DropPath(drop_rate)
        self.drop_path2 = DropPath(drop_rate)
        self.attn = timm_attn(
            dim,
            num_heads=num_heads,
            qkv_bias=False,
            qk_norm=False,
            attn_drop=0.0,
            proj_drop=0.0,
            norm_layer=nn.GELU,
        )
        self.mlp = Mlp(
            in_features=dim,
            hidden_features=int(dim * mlp_ratio),
            act_layer=nn.GELU,
            drop=0.0,
        )

    def forward(self, x):
        x = x + self.drop_path1(self.attn(self.norm1(x)))
        x = x + self.drop_path2(self.mlp(self.norm2(x)))
        return x


class TemporalLayer(nn.Module):
    def __init__(self, dim):
        super().__init__()
        self.transformer_block = zero_module(TransformerBlock(dim))

    def forward(self, x):
        # input: B T C H W
        # output: (B T) C H W
        B, T, C, H, W = x.shape
        x = rearrange(x, "b t c h w -> (b h w) t c")
        x = self.transformer_block(x)
        x = rearrange(x, "(b h w) t c -> (b t) c h w", b=B, h=H, w=W)
        return x


class TemporalLayer2(nn.Module):
    def __init__(self, dim):
        super().__init__()
        self.transformer_block = zero_module(TransformerBlock(dim))

    def forward(self, x, HW):
        # input: (B) (T H W) C
        # output: (B T) C H W
        H, W = HW[0], HW[1]
        x = self.transformer_block(x)
        x = rearrange(x, "(b h w) t c -> (b t) c h w", h=H, w=W)
        return x


class TemporalLayer_noinit(nn.Module):
    def __init__(self, dim):
        super().__init__()
        self.transformer_block = TransformerBlock(dim)

    def forward(self, x, HW):
        # input: (B) (T H W) C
        # output: (B T) C H W
        H, W = HW[0], HW[1]
        x = self.transformer_block(x)
        x = rearrange(x, "(b h w) t c -> (b t) c h w", h=H, w=W)
        return x


class TemporalFlowRegressor(nn.Module):
    def __init__(self, dim):
        super().__init__()
        self.fc1 = nn.Sequential(
            nn.Linear(2, dim),
            nn.GELU(),
        )
        self.transformer_block = zero_module(TransformerBlock(dim))
        self.fc2 = nn.Linear(dim, 2)

    def forward(self, x, HW):
        # input: (B) (T H W) C
        # output: (B T) C H W
        H, W = HW[0], HW[1]
        x = self.fc1(x)
        x = self.transformer_block(x)
        x = self.fc2(x)
        x = rearrange(x, "(b h w) t c -> (b t) c h w", h=H, w=W)
        return x


class TemporalLayer_Ablation(nn.Module):
    def __init__(self, dim):
        super().__init__()
        self.transformer_block = zero_module(TransformerBlock(dim))

    def forward(self, x, HW):
        # input: (B T) (H W) C
        # output: (B T) C H W
        H, W = HW[0], HW[1]
        x = self.transformer_block(x)
        x = rearrange(x, "B (h w) c -> B c h w", h=H, w=W)
        return x


class TemporalFlowHead(nn.Module):
    def __init__(self, C_in, C_out, k_conv):
        super().__init__()
        self.conv_list = nn.ModuleList(
            [
                nn.Conv2d(
                    C_in, C_in, kernel, stride=1, padding=kernel // 2, groups=C_in
                )
                for kernel in k_conv
            ]
        )

        self.ffn1 = nn.Sequential(
            nn.Conv2d(C_in, int(1.5 * C_in), 1, padding=0),
            nn.GELU(),
            nn.Conv2d(int(1.5 * C_in), C_in, 1, padding=0),
        )
        self.pw = nn.Conv2d(C_in, C_in, 1, padding=0)
        self.temporal = TemporalLayer(C_in)
        self.ffn2 = nn.Sequential(
            nn.Conv2d(C_in, int(1.5 * C_in), 1, padding=0),
            nn.GELU(),
            nn.Conv2d(int(1.5 * C_in), C_out, 1, padding=0),
        )

    def forward(self, x):
        # input: B, T, C, H, W
        # output: (B T), C, H, W
        B, T, C, H, W = x.shape
        x = rearrange(x, "b t c h w -> (b t) c h w")
        x = F.gelu(x + self.ffn1(x))
        for conv in self.conv_list:
            x = F.gelu(x + conv(x))
        x = F.gelu(x + self.pw(x))
        x = rearrange(x, "(b t) c h w -> b t c h w", b=B, t=T)
        x = self.temporal(x)
        x = self.ffn2(x)
        return x


class TemporalFlowHead2(nn.Module):
    def __init__(self, C_in, C_out, k_conv):
        super().__init__()
        self.conv_list = nn.ModuleList(
            [
                nn.Conv2d(
                    C_in, C_in, kernel, stride=1, padding=kernel // 2, groups=C_in
                )
                for kernel in k_conv
            ]
        )

        self.ffn1 = nn.Sequential(
            nn.Conv2d(C_in, int(1.5 * C_in), 1, padding=0),
            nn.GELU(),
            nn.Conv2d(int(1.5 * C_in), C_in, 1, padding=0),
        )
        self.pw = nn.Conv2d(C_in, C_in, 1, padding=0)
        self.temporal = nn.Sequential(
            nn.Conv1d(C_in, C_in, 3, padding=1),
            nn.GELU(),
            nn.Conv1d(C_in, C_in, 3, padding=1),
        )
        self.ffn2 = nn.Sequential(
            nn.Conv2d(C_in, int(1.5 * C_in), 1, padding=0),
            nn.GELU(),
            nn.Conv2d(int(1.5 * C_in), C_out, 1, padding=0),
        )

    def forward(self, x):
        # input: B, T, C, H, W
        # output: (B T), C, H, W
        B, T, C, H, W = x.shape
        x = rearrange(x, "b t c h w -> (b t) c h w")
        x = F.gelu(x + self.ffn1(x))
        for conv in self.conv_list:
            x = F.gelu(x + conv(x))
        x = F.gelu(x + self.pw(x))
        # print(rearrange(x, '(b t) c h w -> (b h w) c t', b=B, t=T).shape)
        x = self.temporal(rearrange(x, "(b t) c h w -> (b h w) c t", b=B, t=T))
        x = rearrange(x, "(b h w) c t -> (b t) c h w", h=H, w=W)
        x = self.ffn2(x)
        return x


class SKFlow_Temporal2(nn.Module):
    def __init__(self, args):
        super().__init__()
        self.args = args
        self.encoder = SKMotionEncoder6_Deep_nopool_res(args)
        ratio = 16 if args.Encoder == "UMT" else 8

        self.gma = args.use_gma
        embed_dim = args.decoder_dim // 2

        if self.gma:
            self.aggregator = Aggregate(
                args=self.args,
                dim=embed_dim,
                dim_head=embed_dim,
                heads=self.args.num_heads,
            )
            self.gru = PCBlock4_Deep_nopool_res(
                embed_dim * 4, embed_dim, k_conv=args.PCUpdater_conv
            )
        else:
            self.gru = PCBlock4_Deep_nopool_res(
                embed_dim * 4, embed_dim, k_conv=args.PCUpdater_conv
            )

        self.mask = nn.Sequential(
            nn.Conv2d(embed_dim, embed_dim * 2, 3, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(embed_dim * 2, ratio * ratio * 9, 1, padding=0),
        )

        self.transformer_block = TemporalLayer2(dim=embed_dim)

        self.flow_head = PCBlock4_Deep_nopool_res(embed_dim, 2, args.k_conv)

    def forward(self, nets, inps, corrs, flows, attentions, T=None):
        BT, _, H, W = nets.shape
        B = BT // T

        motion_features = self.encoder(flows, corrs)
        motion_features_globals = self.aggregator(attentions, motion_features)
        motion_features = self.transformer_block(
            rearrange(motion_features, "(B T) C H W -> (B H W) T C", T=T), HW=(H, W)
        )
        inp_cats = torch.cat([inps, motion_features, motion_features_globals], dim=1)

        # Temporal Attention
        # Attentional update
        nets = self.gru(torch.cat([nets, inp_cats], dim=1))
        delta_flows = self.flow_head(nets)

        # scale mask to balence gradients
        masks = 0.25 * self.mask(nets)  # b (t c) h w

        masks = rearrange(masks, "(B T) C H W -> B T C H W", B=B, T=T)
        delta_flows = rearrange(delta_flows, "(B T) C H W -> B T C H W ", B=B, T=T)
        return nets, masks, delta_flows


class SKUpdateBlock_TAM(nn.Module):
    def __init__(self, args):
        super().__init__()
        self.args = args
        self.encoder = SKMotionEncoder6_Deep_nopool_res(args)
        ratio = 16 if args.Encoder == "UMT" else 8

        self.gma = args.use_gma
        embed_dim = args.decoder_dim // 2

        if self.gma:
            self.aggregator = Aggregate(
                args=self.args,
                dim=embed_dim,
                dim_head=embed_dim,
                heads=self.args.num_heads,
            )
            self.gru = PCBlock4_Deep_nopool_res(
                embed_dim * 5, embed_dim, k_conv=args.PCUpdater_conv
            )
        else:
            self.gru = PCBlock4_Deep_nopool_res(
                embed_dim * 5, embed_dim, k_conv=args.PCUpdater_conv
            )

        self.mask = nn.Sequential(
            nn.Conv2d(embed_dim, embed_dim * 2, 3, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(embed_dim * 2, ratio * ratio * 9, 1, padding=0),
        )

        self.transformer_block = TemporalLayer2(dim=embed_dim)

        self.flow_head = PCBlock4_Deep_nopool_res(embed_dim, 2, args.k_conv)

    def forward(self, nets, inps, corrs, flows, attentions, T=None):
        BT, _, H, W = nets.shape
        B = BT // T

        motion_features = self.encoder(flows, corrs)
        motion_features_globals = self.aggregator(attentions, motion_features)
        motion_features_temporal = self.transformer_block(
            rearrange(motion_features, "(B T) C H W -> (B H W) T C", T=T), HW=(H, W)
        )
        inp_cats = torch.cat(
            [inps, motion_features, motion_features_globals, motion_features_temporal],
            dim=1,
        )
        # Temporal Attention
        # Attentional update
        nets = self.gru(torch.cat([nets, inp_cats], dim=1))
        delta_flows = self.flow_head(nets)

        # scale mask to balence gradients
        masks = 0.25 * self.mask(nets)  # b (t c) h w

        masks = rearrange(masks, "(B T) C H W -> B T C H W", B=B, T=T)
        delta_flows = rearrange(delta_flows, "(B T) C H W -> B T C H W ", B=B, T=T)
        return nets, masks, delta_flows


# temporal flow head v1
class SKUpdateBlock_TAM_v3(nn.Module):
    def __init__(
        self,
        decoder_dim,
        num_heads,
        use_gma,
        pcupdater_conv,
        corr_levels,
        corr_radius,
        T,
        k_conv,
    ):
        super().__init__()
        self.encoder = SKMotionEncoder6_Deep_nopool_res(
            decoder_dim=decoder_dim,
            corr_levels=corr_levels,
            corr_radius=corr_radius,
            k_conv=k_conv,
        )
        ratio = 8

        self.gma = use_gma
        embed_dim = decoder_dim // 2

        if self.gma:
            self.aggregator = Aggregate(
                dim=embed_dim, dim_head=embed_dim, heads=num_heads
            )
            self.gru = PCBlock4_Deep_nopool_res(
                embed_dim * 5, embed_dim, k_conv=pcupdater_conv
            )
        else:
            self.gru = PCBlock4_Deep_nopool_res(
                embed_dim * 5, embed_dim, k_conv=pcupdater_conv
            )

        self.mask = nn.Sequential(
            nn.Conv2d(embed_dim, embed_dim * 2, 3, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(embed_dim * 2, ratio * ratio * 9, 1, padding=0),
        )

        self.transformer_block = TemporalLayer2(dim=embed_dim)
        self.flow_head = PCBlock4_Deep_nopool_res(
            embed_dim * (T - 1), 2 * (T - 1), k_conv
        )

    def forward(self, nets, inps, corrs, flows, attentions, T=None):
        BT, _, H, W = nets.shape
        B = BT // T

        motion_features = self.encoder(flows, corrs)
        motion_features_temporal = self.transformer_block(
            rearrange(motion_features, "(B T) C H W -> (B H W) T C", T=T), HW=(H, W)
        )
        if self.gma:
            motion_features_globals = self.aggregator(attentions, motion_features)
            inp_cats = torch.cat(
                [
                    inps,
                    motion_features,
                    motion_features_globals,
                    motion_features_temporal,
                ],
                dim=1,
            )
        else:
            inp_cats = torch.cat(
                [inps, motion_features, motion_features_temporal],
                dim=1,
            )
        # Attentional update
        nets = self.gru(torch.cat([nets, inp_cats], dim=1))
        delta_flows = self.flow_head(
            rearrange(nets, "(B T) C H W -> B (T C) H W", T=T)
        )  # (b t) c h w => b (t c) h w

        # scale mask to balence gradients
        masks = 0.25 * self.mask(nets)

        masks = rearrange(masks, "(B T) C H W -> B T C H W", B=B, T=T)
        delta_flows = rearrange(delta_flows, "B (T C) H W -> B T C H W ", T=T)

        return nets, masks, delta_flows


# temporal flow head v1
class SKUpdateBlock_TAM_v3_noinit(nn.Module):
    def __init__(self, args):
        super().__init__()
        self.args = args
        self.encoder = SKMotionEncoder6_Deep_nopool_res(args)
        ratio = 16 if args.Encoder == "UMT" else 8

        self.gma = args.use_gma
        embed_dim = args.decoder_dim // 2

        if self.gma:
            self.aggregator = Aggregate(
                args=self.args,
                dim=embed_dim,
                dim_head=embed_dim,
                heads=self.args.num_heads,
            )
            self.gru = PCBlock4_Deep_nopool_res(
                embed_dim * 5, embed_dim, k_conv=args.PCUpdater_conv
            )
        else:
            self.gru = PCBlock4_Deep_nopool_res(
                embed_dim * 5, embed_dim, k_conv=args.PCUpdater_conv
            )

        self.mask = nn.Sequential(
            nn.Conv2d(embed_dim, embed_dim * 2, 3, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(embed_dim * 2, ratio * ratio * 9, 1, padding=0),
        )

        self.transformer_block = TemporalLayer_noinit(dim=embed_dim)
        self.flow_head = PCBlock4_Deep_nopool_res(
            embed_dim * (args.T - 1), 2 * (args.T - 1), args.k_conv
        )

    def forward(self, nets, inps, corrs, flows, attentions, T=None):
        BT, _, H, W = nets.shape
        B = BT // T

        motion_features = self.encoder(flows, corrs)
        motion_features_globals = self.aggregator(attentions, motion_features)
        motion_features_temporal = self.transformer_block(
            rearrange(motion_features, "(B T) C H W -> (B H W) T C", T=T), HW=(H, W)
        )
        inp_cats = torch.cat(
            [inps, motion_features, motion_features_globals, motion_features_temporal],
            dim=1,
        )
        # Temporal Attention
        # Attentional update
        nets = self.gru(torch.cat([nets, inp_cats], dim=1))
        # print(T, rearrange(nets, '(B T) C H W -> B (T C) H W', T=T).shape, nets.shape)
        delta_flows = self.flow_head(
            rearrange(nets, "(B T) C H W -> B (T C) H W", T=T)
        )  # (b t) c h w => b (t c) h w

        # scale mask to balence gradients
        masks = 0.25 * self.mask(nets)

        masks = rearrange(masks, "(B T) C H W -> B T C H W", B=B, T=T)
        delta_flows = rearrange(delta_flows, "B (T C) H W -> B T C H W ", T=T)

        return nets, masks, delta_flows


# temporal flow head v1
class SKUpdateBlock_TAM_ParamAblation(nn.Module):
    def __init__(self, args):
        super().__init__()
        self.args = args
        self.encoder = SKMotionEncoder6_Deep_nopool_res(args)
        ratio = 16 if args.Encoder == "UMT" else 8

        self.gma = args.use_gma
        embed_dim = args.decoder_dim // 2

        if self.gma:
            self.aggregator = Aggregate(
                args=self.args,
                dim=embed_dim,
                dim_head=embed_dim,
                heads=self.args.num_heads,
            )
            self.gru = PCBlock4_Deep_nopool_res(
                embed_dim * 5, embed_dim, k_conv=args.PCUpdater_conv
            )
        else:
            self.gru = PCBlock4_Deep_nopool_res(
                embed_dim * 5, embed_dim, k_conv=args.PCUpdater_conv
            )

        self.mask = nn.Sequential(
            nn.Conv2d(embed_dim, embed_dim * 2, 3, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(embed_dim * 2, ratio * ratio * 9, 1, padding=0),
        )

        self.transformer_block = TemporalLayer_Ablation(dim=embed_dim)
        self.flow_head = PCBlock4_Deep_nopool_res(
            embed_dim * (args.T - 1), 2 * (args.T - 1), args.k_conv
        )

    def forward(self, nets, inps, corrs, flows, attentions, T=None):
        BT, _, H, W = nets.shape
        B = BT // T

        motion_features = self.encoder(flows, corrs)
        motion_features_globals = self.aggregator(attentions, motion_features)
        motion_features_additional = self.transformer_block(
            rearrange(motion_features, "(B T) C H W -> (B T) (H W) C", T=T), HW=[H, W]
        )
        inp_cats = torch.cat(
            [
                inps,
                motion_features,
                motion_features_globals,
                motion_features_additional,
            ],
            dim=1,
        )
        # Temporal Attention
        # Attentional update
        nets = self.gru(torch.cat([nets, inp_cats], dim=1))
        delta_flows = self.flow_head(
            rearrange(nets, "(B T) C H W -> B (T C) H W", B=B)
        )  # (b t) c h w => b (t c) h w

        # scale mask to balence gradients
        masks = 0.25 * self.mask(nets)

        masks = rearrange(masks, "(B T) C H W -> B T C H W", B=B, T=T)
        delta_flows = rearrange(delta_flows, "B (T C) H W -> B T C H W ", T=T)

        return nets, masks, delta_flows


# temporal flow head v1
class SKUpdateBlock_TAM_v7(nn.Module):
    def __init__(self, args):
        super().__init__()
        self.args = args
        self.encoder = SKMotionEncoder6_Deep_nopool_res(args)
        ratio = 16 if args.Encoder == "UMT" else 8

        self.gma = args.use_gma
        embed_dim = args.decoder_dim // 2

        if self.gma:
            self.aggregator = Aggregate(
                args=self.args,
                dim=embed_dim,
                dim_head=embed_dim,
                heads=self.args.num_heads,
            )
            self.gru = PCBlock4_Deep_nopool_res(
                embed_dim * 5, embed_dim, k_conv=args.PCUpdater_conv
            )
        else:
            self.gru = PCBlock4_Deep_nopool_res(
                embed_dim * 5, embed_dim, k_conv=args.PCUpdater_conv
            )

        self.mask = nn.Sequential(
            nn.Conv2d(embed_dim, embed_dim * 2, 3, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(embed_dim * 2, ratio * ratio * 9, 1, padding=0),
        )

        self.transformer_block = TemporalLayer2(dim=embed_dim)
        self.temporal_regressor = TemporalLayer2(dim=2)
        self.flow_head = PCBlock4_Deep_nopool_res(
            embed_dim * (args.T - 1), 2 * (args.T - 1), args.k_conv
        )

    def forward(self, nets, inps, corrs, flows, attentions, T=None):
        BT, _, H, W = nets.shape
        B = BT // T

        motion_features = self.encoder(flows, corrs)
        motion_features_globals = self.aggregator(attentions, motion_features)
        motion_features_temporal = self.transformer_block(
            rearrange(motion_features, "(B T) C H W -> (B H W) T C", T=T), HW=(H, W)
        )
        inp_cats = torch.cat(
            [inps, motion_features, motion_features_globals, motion_features_temporal],
            dim=1,
        )
        # Temporal Attention
        # Attentional update
        nets = self.gru(torch.cat([nets, inp_cats], dim=1))
        delta_flows = self.flow_head(
            rearrange(nets, "(B T) C H W -> B (T C) H W", B=B)
        )  # (b t) c h w => b (t c) h w

        # scale mask to balence gradients
        masks = 0.25 * self.mask(nets)

        masks = rearrange(masks, "(B T) C H W -> B T C H W", B=B, T=T)
        delta_flows = rearrange(delta_flows, "B (T C) H W -> (B H W) T C ", T=T)
        delta_flows = self.temporal_regressor(delta_flows, HW=(H, W))
        delta_flows = rearrange(delta_flows, "(B T) C H W -> B T C H W", B=B)

        return nets, masks, delta_flows


# temporal flow head v1
class SKUpdateBlock_TAM_v8(nn.Module):
    def __init__(self, args):
        super().__init__()
        self.args = args
        self.encoder = SKMotionEncoder6_Deep_nopool_res(args)
        ratio = 16 if args.Encoder == "UMT" else 8

        self.gma = args.use_gma
        embed_dim = args.decoder_dim // 2

        if self.gma:
            self.aggregator = Aggregate(
                args=self.args,
                dim=embed_dim,
                dim_head=embed_dim,
                heads=self.args.num_heads,
            )
            self.gru = PCBlock4_Deep_nopool_res(
                embed_dim * 5, embed_dim, k_conv=args.PCUpdater_conv
            )
        else:
            self.gru = PCBlock4_Deep_nopool_res(
                embed_dim * 5, embed_dim, k_conv=args.PCUpdater_conv
            )

        self.mask = nn.Sequential(
            nn.Conv2d(embed_dim, embed_dim * 2, 3, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(embed_dim * 2, ratio * ratio * 9, 1, padding=0),
        )

        self.transformer_block = TemporalLayer2(dim=embed_dim)
        self.temporal_regressor = TemporalFlowRegressor(dim=embed_dim)
        self.flow_head = PCBlock4_Deep_nopool_res(
            embed_dim * (args.T - 1), 2 * (args.T - 1), args.k_conv
        )

    def forward(self, nets, inps, corrs, flows, attentions, T=None):
        BT, _, H, W = nets.shape
        B = BT // T

        motion_features = self.encoder(flows, corrs)
        motion_features_globals = self.aggregator(attentions, motion_features)
        motion_features_temporal = self.transformer_block(
            rearrange(motion_features, "(B T) C H W -> (B H W) T C", T=T), HW=(H, W)
        )
        inp_cats = torch.cat(
            [inps, motion_features, motion_features_globals, motion_features_temporal],
            dim=1,
        )
        # Temporal Attention
        # Attentional update
        nets = self.gru(torch.cat([nets, inp_cats], dim=1))
        delta_flows = self.flow_head(
            rearrange(nets, "(B T) C H W -> B (T C) H W", B=B)
        )  # (b t) c h w => b (t c) h w

        # scale mask to balence gradients
        masks = 0.25 * self.mask(nets)

        masks = rearrange(masks, "(B T) C H W -> B T C H W", B=B, T=T)
        delta_flows = rearrange(delta_flows, "B (T C) H W -> (B H W) T C ", T=T)
        delta_flows = self.temporal_regressor(delta_flows, HW=(H, W))
        delta_flows = rearrange(delta_flows, "(B T) C H W -> B T C H W", B=B)

        return nets, masks, delta_flows


# temporal flow head v2
class SKUpdateBlock_TAM_v4(nn.Module):
    def __init__(self, args):
        super().__init__()
        self.args = args
        self.encoder = SKMotionEncoder6_Deep_nopool_res(args)
        ratio = 16 if args.Encoder == "UMT" else 8

        self.gma = args.use_gma
        embed_dim = args.decoder_dim // 2

        if self.gma:
            self.aggregator = Aggregate(
                args=self.args,
                dim=embed_dim,
                dim_head=embed_dim,
                heads=self.args.num_heads,
            )
            self.gru = PCBlock4_Deep_nopool_res(
                embed_dim * 5, embed_dim, k_conv=args.PCUpdater_conv
            )
        else:
            self.gru = PCBlock4_Deep_nopool_res(
                embed_dim * 5, embed_dim, k_conv=args.PCUpdater_conv
            )

        self.mask = nn.Sequential(
            nn.Conv2d(embed_dim, embed_dim * 2, 3, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(embed_dim * 2, ratio * ratio * 9, 1, padding=0),
        )

        self.transformer_block = TemporalLayer2(dim=embed_dim)

        self.transformer_block2 = TemporalLayer2(dim=embed_dim)

        self.flow_head = PCBlock4_Deep_nopool_res(embed_dim, 2, args.k_conv)

    def forward(self, nets, inps, corrs, flows, attentions, T=None):
        BT, _, H, W = nets.shape
        B = BT // T

        motion_features = self.encoder(flows, corrs)
        motion_features_globals = self.aggregator(attentions, motion_features)
        motion_features_temporal = self.transformer_block(
            rearrange(motion_features, "(B T) C H W -> (B H W) T C", T=T), HW=(H, W)
        )
        inp_cats = torch.cat(
            [inps, motion_features, motion_features_globals, motion_features_temporal],
            dim=1,
        )
        # Temporal Attention
        # Attentional update
        nets = self.gru(torch.cat([nets, inp_cats], dim=1))
        nets = self.transformer_block2(
            rearrange(nets, "(B T) C H W -> (B H W) T C", T=T), HW=(H, W)
        )
        delta_flows = self.flow_head(nets)

        # scale mask to balence gradients
        masks = 0.25 * self.mask(nets)  # b (t c) h w

        masks = rearrange(masks, "(B T) C H W -> B T C H W", B=B, T=T)
        delta_flows = rearrange(delta_flows, "(B T) C H W -> B T C H W ", B=B, T=T)
        return nets, masks, delta_flows


# temporal flow head v1 + mask
class SKUpdateBlock_TAM_v5(nn.Module):
    def __init__(self, args):
        super().__init__()
        self.args = args
        self.encoder = SKMotionEncoder6_Deep_nopool_res(args)
        ratio = 16 if args.Encoder == "UMT" else 8

        self.gma = args.use_gma
        embed_dim = args.decoder_dim // 2

        if self.gma:
            self.aggregator = Aggregate(
                args=self.args,
                dim=embed_dim,
                dim_head=embed_dim,
                heads=self.args.num_heads,
            )
            self.gru = PCBlock4_Deep_nopool_res(
                embed_dim * 5, embed_dim, k_conv=args.PCUpdater_conv
            )
        else:
            self.gru = PCBlock4_Deep_nopool_res(
                embed_dim * 5, embed_dim, k_conv=args.PCUpdater_conv
            )

        self.mask = nn.Sequential(
            nn.Conv2d(embed_dim * (args.T - 1), embed_dim * 2, 3, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(embed_dim * 2, ratio * ratio * 9 * (args.T - 1), 1, padding=0),
        )

        self.transformer_block = TemporalLayer2(dim=embed_dim)

        self.flow_head = PCBlock4_Deep_nopool_res(
            embed_dim * (args.T - 1), 2 * (args.T - 1), args.k_conv
        )

    def forward(self, nets, inps, corrs, flows, attentions, T=None):
        BT, _, H, W = nets.shape
        B = BT // T

        motion_features = self.encoder(flows, corrs)
        motion_features_globals = self.aggregator(attentions, motion_features)
        motion_features_temporal = self.transformer_block(
            rearrange(motion_features, "(B T) C H W -> (B H W) T C", T=T), HW=(H, W)
        )
        inp_cats = torch.cat(
            [inps, motion_features, motion_features_globals, motion_features_temporal],
            dim=1,
        )
        # Temporal Attention
        # Attentional update
        nets = self.gru(torch.cat([nets, inp_cats], dim=1))
        delta_flows = self.flow_head(
            rearrange(nets, "(B T) C H W -> B (T C) H W", B=B)
        )  # (b t) c h w => b (t c) h w

        # scale mask to balence gradients
        masks = 0.25 * self.mask(rearrange(nets, "(B T) C H W -> B (T C) H W", B=B))

        masks = rearrange(masks, "B (T C) H W -> B T C H W", T=T)
        delta_flows = rearrange(delta_flows, "B (T C) H W -> B T C H W ", T=T)

        return nets, masks, delta_flows


# temporal flow head v1 + corr temporal
class SKUpdateBlock_TAM_v6(nn.Module):
    def __init__(self, args):
        super().__init__()
        self.args = args
        self.encoder = SKMotionEncoder_CorrTemporal(args)
        ratio = 16 if args.Encoder == "UMT" else 8

        self.gma = args.use_gma
        embed_dim = args.decoder_dim // 2

        if self.gma:
            self.aggregator = Aggregate(
                args=self.args,
                dim=embed_dim,
                dim_head=embed_dim,
                heads=self.args.num_heads,
            )
            self.gru = PCBlock4_Deep_nopool_res(
                embed_dim * 5, embed_dim, k_conv=args.PCUpdater_conv
            )
        else:
            self.gru = PCBlock4_Deep_nopool_res(
                embed_dim * 5, embed_dim, k_conv=args.PCUpdater_conv
            )

        self.mask = nn.Sequential(
            nn.Conv2d(embed_dim, embed_dim * 2, 3, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(embed_dim * 2, ratio * ratio * 9, 1, padding=0),
        )

        self.transformer_block = TemporalLayer2(dim=embed_dim)

        self.flow_head = PCBlock4_Deep_nopool_res(
            embed_dim * (args.T - 1), 2 * (args.T - 1), args.k_conv
        )

    def forward(self, nets, inps, corrs, flows, attentions, T=None):
        BT, _, H, W = nets.shape
        B = BT // T

        motion_features = self.encoder(flows, corrs)
        motion_features_globals = self.aggregator(attentions, motion_features)
        motion_features_temporal = self.transformer_block(
            rearrange(motion_features, "(B T) C H W -> (B H W) T C", T=T), HW=(H, W)
        )
        inp_cats = torch.cat(
            [inps, motion_features, motion_features_globals, motion_features_temporal],
            dim=1,
        )
        # Temporal Attention
        # Attentional update
        nets = self.gru(torch.cat([nets, inp_cats], dim=1))
        delta_flows = self.flow_head(
            rearrange(nets, "(B T) C H W -> B (T C) H W", B=B)
        )  # (b t) c h w => b (t c) h w

        # scale mask to balence gradients
        masks = 0.25 * self.mask(nets)

        masks = rearrange(masks, "(B T) C H W -> B T C H W", B=B, T=T)
        delta_flows = rearrange(delta_flows, "B (T C) H W -> B T C H W ", T=T)

        return nets, masks, delta_flows


class Bi_SKMotionEncoder6_Deep_nopool_res(nn.Module):
    def __init__(self, args):
        super().__init__()
        self.args = args
        self.T = args.T - 1
        out_dim = args.decoder_dim // 2
        cor_planes = args.corr_levels * (2 * args.corr_radius + 1) ** 2

        self.convc1 = PCBlock4_Deep_nopool_res(
            cor_planes, int(256 / self.T), args.k_conv
        )
        self.convc2 = PCBlock4_Deep_nopool_res(256, 192, args.k_conv)

        self.convf1 = nn.Conv2d(2 * self.T, 128, 1, 1, 0)
        self.convf2 = PCBlock4_Deep_nopool_res(128, 64, args.k_conv)

        self.conv = PCBlock4_Deep_nopool_res(
            64 + 192, out_dim - 2 * self.T, args.k_conv
        )

    def forward(self, flow, corr, attention=None):
        sub_corrs = list(corr.chunk(self.T, dim=1))
        sub_corrs = [self.convc1(cor) for cor in sub_corrs]

        cor = F.gelu(torch.cat(sub_corrs, dim=1))
        cor = self.convc2(cor)

        flo = self.convf1(flow)
        flo = self.convf2(flo)

        cor_flo = torch.cat([cor, flo], dim=1)
        out = self.conv(cor_flo)

        return torch.cat([out, flow], dim=1)


class Bi_SKUpdateBlock_TAM(nn.Module):
    def __init__(self, args):
        super().__init__()
        self.args = args
        self.encoder = Bi_SKMotionEncoder6_Deep_nopool_res(args)
        ratio = 16 if args.Encoder == "UMT" else 8

        self.gma = args.use_gma
        embed_dim = args.decoder_dim // 2

        if self.gma:
            self.aggregator = Aggregate(
                args=self.args,
                dim=embed_dim,
                dim_head=embed_dim,
                heads=self.args.num_heads,
            )
            self.gru = PCBlock4_Deep_nopool_res(
                embed_dim * 5, embed_dim, k_conv=args.PCUpdater_conv
            )
        else:
            self.gru = PCBlock4_Deep_nopool_res(
                embed_dim * 5, embed_dim, k_conv=args.PCUpdater_conv
            )

        self.mask = nn.Sequential(
            nn.Conv2d(embed_dim, embed_dim * 2, 3, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(embed_dim * 2, ratio * ratio * 9 * (args.T - 1), 1, padding=0),
        )

        self.transformer_block = TemporalLayer2(dim=embed_dim)

        self.flow_head = PCBlock4_Deep_nopool_res(
            embed_dim, 2 * (args.T - 1), args.k_conv
        )

    def forward(self, nets, inps, corrs, flows, attentions, T=None):
        BT, _, H, W = nets.shape
        B = BT // T

        motion_features = self.encoder(flows, corrs)
        motion_features_globals = self.aggregator(attentions, motion_features)
        motion_features_temporal = self.transformer_block(
            rearrange(motion_features, "(B T) C H W -> (B H W) T C", T=T), HW=(H, W)
        )
        inp_cats = torch.cat(
            [inps, motion_features, motion_features_globals, motion_features_temporal],
            dim=1,
        )
        # Temporal Attention
        # Attentional update
        nets = self.gru(torch.cat([nets, inp_cats], dim=1))
        delta_flows = self.flow_head(nets)

        # scale mask to balence gradients
        masks = 0.25 * self.mask(nets)

        masks = rearrange(masks, "(B T) C H W -> B T C H W", B=B, T=T)
        delta_flows = rearrange(delta_flows, "(B T) C H W -> B T C H W ", B=B, T=T)
        return nets, masks, delta_flows


class SKFlow_TMM(nn.Module):
    def __init__(self, args):
        super().__init__()
        self.args = args
        self.encoder = SKMotionEncoder6_Deep_nopool_res(args)
        ratio = 16 if args.Encoder == "UMT" else 8

        self.gma = args.use_gma
        embed_dim = args.decoder_dim // 2

        if self.gma:
            self.aggregator = TMMAggregate(
                args=self.args,
                dim=embed_dim,
                dim_head=embed_dim,
                heads=self.args.num_heads,
            )
            self.gru = PCBlock4_Deep_nopool_res(
                embed_dim * 4, embed_dim, k_conv=args.PCUpdater_conv
            )
        else:
            self.gru = PCBlock4_Deep_nopool_res(
                embed_dim * 4, embed_dim, k_conv=args.PCUpdater_conv
            )

        self.mask = nn.Sequential(
            nn.Conv2d(embed_dim, embed_dim * 2, 3, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(embed_dim * 2, ratio * ratio * 9, 1, padding=0),
        )

        self.flow_head = PCBlock4_Deep_nopool_res(embed_dim, 2, args.k_conv)

    def forward(
        self, nets, inps, corrs, flows, attentions, temporal_attentions, T=None
    ):
        BT, _, H, W = nets.shape
        B = BT // T

        motion_features = self.encoder(flows, corrs)
        motion_features_globals = self.aggregator(
            attentions, temporal_attentions, motion_features
        )
        # motion_features = self.transformer_block(rearrange(motion_features, '(B T) C H W -> (B H W) T C', T=T), HW=(H, W))
        inp_cats = torch.cat([inps, motion_features, motion_features_globals], dim=1)

        # Temporal Attention
        # Attentional update
        nets = self.gru(torch.cat([nets, inp_cats], dim=1))
        delta_flows = self.flow_head(nets)

        # scale mask to balence gradients
        masks = 0.25 * self.mask(nets)

        masks = rearrange(masks, "(B T) C H W -> B T C H W", B=B, T=T)
        delta_flows = rearrange(delta_flows, "(B T) C H W -> B T C H W ", B=B, T=T)
        return nets, masks, delta_flows


# TODO
class CrossTransformerBlock(nn.Module):
    def __init__(self, in_channels=128):
        super().__init__()
        self.q_proj = nn.Linear(in_channels, in_channels)
        self.k_proj = nn.Linear(in_channels, in_channels)

        for p in self.parameters():
            if p.dim() > 1:
                nn.init.xavier_uniform_(p)

    def forward(self, flow, feature):
        b, c, h, w = feature.size()
        # f1为v; f2为q, k
        query = feature.view(b, c, h * w).permute(0, 2, 1)
        query = self.q_proj(query)  # [B, H*W, C]
        key = self.k_proj(query)  # [B, H*W, C]
        value = flow.view(b, flow.size(1), h * w).permute(0, 2, 1)  # [B, H*W, 2]

        scores = torch.matmul(query, key.permute(0, 2, 1)) / (c**0.5)  # [B, H*W, H*W]
        prob = torch.softmax(scores, dim=-1)

        out = torch.matmul(prob, value)  # [B, H*W, 2]
        out = out.view(b, h, w, value.size(-1)).permute(0, 3, 1, 2)
        return out


class FlowRegressorv1(nn.Module):
    def __init__(self):
        super().__init__()
        self.conv1 = nn.Conv2d(2, 128, 3, padding=1)
        # self.conv2 = nn.Conv2d(128, 2, 3, padding=1)
        self.cross_attn_layer = CrossTransformerBlock()

    def forward(self, flo1, flo2):
        feat1 = self.conv1(flo1)
        feat2 = self.conv1(flo2)

        res1 = self.corss_attn_layer(flo1, feat2)
        res2 = self.corss_attn_layer(flo2, feat1)
        flo1 = flo1 + res1
        flo2 = flo2 + res2
        flows = torch.stack([flo1, flo2], dim=1)
        flows = rearrange(flows, "B T C H W -> (B T) C H W")
        # flows = self.conv2(flows)

        return flows


class SKFlow_TMM3(nn.Module):
    def __init__(self, args):
        super().__init__()
        self.args = args
        self.encoder = SKMotionEncoder6_Deep_nopool_res(args)
        ratio = 16 if args.Encoder == "UMT" else 8

        self.gma = args.use_gma
        embed_dim = args.decoder_dim // 2

        if self.gma:
            self.aggregator = TMMAggregate(
                args=self.args,
                dim=embed_dim,
                dim_head=embed_dim,
                heads=self.args.num_heads,
            )
            self.gru = PCBlock4_Deep_nopool_res(
                embed_dim * 4, embed_dim, k_conv=args.PCUpdater_conv
            )
        else:
            self.gru = PCBlock4_Deep_nopool_res(
                embed_dim * 4, embed_dim, k_conv=args.PCUpdater_conv
            )

        self.mask = nn.Sequential(
            nn.Conv2d(embed_dim, embed_dim * 2, 3, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(embed_dim * 2, ratio * ratio * 9, 1, padding=0),
        )

        self.flow_head = PCBlock4_Deep_nopool_res(embed_dim, 2, args.k_conv)
        self.flow_regressor = FlowRegressorv1()

    def forward(
        self, nets, inps, corrs, flows, attentions, temporal_attentions, T=None
    ):
        BT, _, H, W = nets.shape
        B = BT // T

        motion_features = self.encoder(flows, corrs)
        motion_features_globals = self.aggregator(
            attentions, temporal_attentions, motion_features
        )
        # motion_features = self.transformer_block(rearrange(motion_features, '(B T) C H W -> (B H W) T C', T=T), HW=(H, W))
        inp_cats = torch.cat([inps, motion_features, motion_features_globals], dim=1)

        # Temporal Attention
        # Attentional update
        nets = self.gru(torch.cat([nets, inp_cats], dim=1))
        delta_flows = self.flow_head(nets)  # [(B, T) C H W]

        # scale mask to balence gradients
        masks = 0.25 * self.mask(nets)

        masks = rearrange(masks, "(B T) C H W -> B T C H W", B=B, T=T)
        delta_flows = rearrange(delta_flows, "(B T) C H W -> B T C H W", B=B, T=T)
        delta_flows = self.flow_regressor(
            delta_flows[:, 0, ...], delta_flows[:, 1, ...]
        )

        return nets, masks, delta_flows


# temporal attn layer
class SKFlow_TMM2(nn.Module):
    def __init__(self, args):
        super().__init__()
        self.args = args
        self.encoder = SKMotionEncoder6_Deep_nopool_res(args)
        ratio = 16 if args.Encoder == "UMT" else 8

        self.gma = args.use_gma
        embed_dim = args.decoder_dim // 2

        if self.gma:
            self.aggregator = TMMAggregate(
                args=self.args,
                dim=embed_dim,
                dim_head=embed_dim,
                heads=self.args.num_heads,
            )
            self.gru = PCBlock4_Deep_nopool_res(
                embed_dim * 4, embed_dim, k_conv=args.PCUpdater_conv
            )
        else:
            self.gru = PCBlock4_Deep_nopool_res(
                embed_dim * 4, embed_dim, k_conv=args.PCUpdater_conv
            )

        self.mask = nn.Sequential(
            nn.Conv2d(embed_dim, embed_dim * 2, 3, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(embed_dim * 2, ratio * ratio * 9, 1, padding=0),
        )

        self.flow_head = TemporalFlowHead(embed_dim, 2, args.k_conv)
        # self.flow_head = TemporalFlowHead2(embed_dim, 2, args.k_conv)
        # self.flow_head = nn.Sequential(
        #     PCBlock4_Deep_nopool_res(embed_dim, embed_dim//2, args.k_conv),
        #     SKBlock_Temporal(embed_dim//2, 2, args.k_conv),
        # )

    def forward(
        self, nets, inps, corrs, flows, attentions, temporal_attentions, T=None
    ):
        BT, _, H, W = nets.shape
        B = BT // T

        motion_features = self.encoder(flows, corrs)
        motion_features_globals = self.aggregator(
            attentions, temporal_attentions, motion_features
        )
        # motion_features = self.transformer_block(rearrange(motion_features, '(B T) C H W -> (B H W) T C', T=T), HW=(H, W))
        inp_cats = torch.cat([inps, motion_features, motion_features_globals], dim=1)

        # Temporal Attention
        # Attentional update
        nets = self.gru(torch.cat([nets, inp_cats], dim=1))
        delta_flows = self.flow_head(rearrange(nets, "(B T) C H W -> B T C H W", T=T))
        # delta_flows = self.flow_head(nets)

        # scale mask to balence gradients
        masks = 0.25 * self.mask(nets)

        masks = rearrange(masks, "(B T) C H W -> B T C H W", B=B, T=T)
        delta_flows = rearrange(delta_flows, "(B T) C H W -> B T C H W ", B=B, T=T)
        return nets, masks, delta_flows


# temporal sk


class SKFlow_TMM3(nn.Module):
    def __init__(self, args):
        super().__init__()
        self.args = args
        self.encoder = SKMotionEncoder6_Deep_nopool_res(args)
        ratio = 16 if args.Encoder == "UMT" else 8

        self.gma = args.use_gma
        embed_dim = args.decoder_dim // 2

        if self.gma:
            self.aggregator = TMMAggregate(
                args=self.args,
                dim=embed_dim,
                dim_head=embed_dim,
                heads=self.args.num_heads,
            )
            self.gru = PCBlock4_Deep_nopool_res(
                embed_dim * 4, embed_dim, k_conv=args.PCUpdater_conv
            )
        else:
            self.gru = PCBlock4_Deep_nopool_res(
                embed_dim * 4, embed_dim, k_conv=args.PCUpdater_conv
            )

        self.mask = nn.Sequential(
            nn.Conv2d(embed_dim, embed_dim * 2, 3, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(embed_dim * 2, ratio * ratio * 9, 1, padding=0),
        )

        # self.flow_head = TemporalFlowHead(embed_dim, 2, args.k_conv)
        # self.flow_head = TemporalFlowHead2(embed_dim, 2, args.k_conv)
        self.flow_head = nn.Sequential(
            PCBlock4_Deep_nopool_res(embed_dim, embed_dim // 2, args.k_conv),
            SKBlock_Temporal(embed_dim // 2, 2, args.k_conv),
        )

    def forward(
        self, nets, inps, corrs, flows, attentions, temporal_attentions, T=None
    ):
        BT, _, H, W = nets.shape
        B = BT // T

        motion_features = self.encoder(flows, corrs)
        motion_features_globals = self.aggregator(
            attentions, temporal_attentions, motion_features
        )
        # motion_features = self.transformer_block(rearrange(motion_features, '(B T) C H W -> (B H W) T C', T=T), HW=(H, W))
        inp_cats = torch.cat([inps, motion_features, motion_features_globals], dim=1)

        # Temporal Attention
        # Attentional update
        nets = self.gru(torch.cat([nets, inp_cats], dim=1))
        # delta_flows = self.flow_head(rearrange(nets, '(B T) C H W -> B T C H W', T=T))
        delta_flows = self.flow_head(nets)

        # scale mask to balence gradients
        masks = 0.25 * self.mask(nets)

        masks = rearrange(masks, "(B T) C H W -> B T C H W", B=B, T=T)
        delta_flows = rearrange(delta_flows, "(B T) C H W -> B T C H W ", B=B, T=T)
        return nets, masks, delta_flows


class SKFlow_Temporal3(nn.Module):
    def __init__(self, args):
        super().__init__()
        self.args = args
        self.encoder = SKMotionEncoder6_Deep_nopool_res(args)
        ratio = 16 if args.Encoder == "UMT" else 8

        self.gma = args.use_gma
        embed_dim = args.decoder_dim // 2

        if self.gma:
            self.aggregator = Aggregate(
                args=self.args,
                dim=embed_dim,
                dim_head=embed_dim,
                heads=self.args.num_heads,
            )
            self.gru = PCBlock4_Deep_nopool_res(
                embed_dim * 4, embed_dim, k_conv=args.PCUpdater_conv
            )
        else:
            self.gru = PCBlock4_Deep_nopool_res(
                embed_dim * 4, embed_dim, k_conv=args.PCUpdater_conv
            )

        self.mask = nn.Sequential(
            nn.Conv2d(embed_dim, embed_dim * 2, 3, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(embed_dim * 2, ratio * ratio * 9, 1, padding=0),
        )

        self.skblock = PCBlock4_Deep_nopool_res(
            embed_dim, embed_dim, k_conv=args.k_conv
        )

        self.flow_head = PCBlock4_Deep_nopool_res(embed_dim, 2, args.k_conv)

    def forward(self, nets, inps, corrs, flows, attentions, T=None):
        BT, _, H, W = nets.shape
        B = BT // T

        motion_features = self.encoder(flows, corrs)
        motion_features_globals = self.aggregator(attentions, motion_features)
        motion_features = self.skblock(
            rearrange(motion_features, "(B T) C H W -> B C H (T W)", T=T)
        )
        motion_features = rearrange(motion_features, "B C H (T W) -> (B T) C H W", T=T)
        inp_cats = torch.cat([inps, motion_features, motion_features_globals], dim=1)

        # Temporal Attention
        # Attentional update
        nets = self.gru(torch.cat([nets, inp_cats], dim=1))
        delta_flows = self.flow_head(nets)

        # scale mask to balence gradients
        masks = 0.25 * self.mask(nets)  # b (t c) h w

        masks = rearrange(masks, "(B T) C H W -> B T C H W", B=B, T=T)
        delta_flows = rearrange(delta_flows, "(B T) C H W -> B T C H W ", B=B, T=T)
        return nets, masks, delta_flows


class SKFlow_Temporal4(nn.Module):
    def __init__(self, args):
        super().__init__()
        self.args = args
        self.encoder = SKMotionEncoder6_Deep_nopool_res(args)
        ratio = 16 if args.Encoder == "UMT" else 8

        self.gma = args.use_gma
        embed_dim = args.decoder_dim // 2

        if self.gma:
            self.aggregator = Aggregate(
                args=self.args,
                dim=embed_dim,
                dim_head=embed_dim,
                heads=self.args.num_heads,
            )
            self.gru = PCBlock4_Deep_nopool_res(
                embed_dim * 4, embed_dim, k_conv=args.PCUpdater_conv
            )
        else:
            self.gru = PCBlock4_Deep_nopool_res(
                embed_dim * 4, embed_dim, k_conv=args.PCUpdater_conv
            )

        self.mask = nn.Sequential(
            nn.Conv2d(embed_dim, embed_dim * 2, 3, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(embed_dim * 2, ratio * ratio * 9, 1, padding=0),
        )

        self.skblock = PCBlock4_Deep_nopool_res(
            embed_dim, embed_dim, k_conv=args.k_conv
        )

        self.flow_head = PCBlock4_Deep_nopool_res(embed_dim, 2, args.k_conv)

    def forward(self, nets, inps, corrs, flows, attentions, T=None):
        BT, _, H, W = nets.shape
        B = BT // T

        motion_features = self.encoder(flows, corrs)
        motion_features_globals = self.aggregator(attentions, motion_features)
        motion_features = self.skblock(
            rearrange(motion_features, "(B T) C H W -> B C H (T W)", T=T)
        )
        motion_features = rearrange(motion_features, "B C H (T W) -> (B T) C H W", T=T)
        inp_cats = torch.cat([inps, motion_features, motion_features_globals], dim=1)

        # Temporal Attention
        # Attentional update
        nets = self.gru(torch.cat([nets, inp_cats], dim=1))
        delta_flows = self.flow_head(nets)

        # scale mask to balence gradients
        masks = 0.25 * self.mask(nets)  # b (t c) h w

        masks = rearrange(masks, "(B T) C H W -> B T C H W", B=B, T=T)
        delta_flows = rearrange(delta_flows, "(B T) C H W -> B T C H W ", B=B, T=T)
        return nets, masks, delta_flows


class SKFlow_Temporal(nn.Module):
    def __init__(self, args):
        super().__init__()
        self.args = args
        self.encoder = SKMotionEncoder6_Deep_nopool_res(args)
        ratio = 16 if args.Encoder == "UMT" else 8

        self.gma = args.use_gma
        embed_dim = args.decoder_dim // 2

        if self.gma:
            self.aggregator = Aggregate(
                args=self.args,
                dim=embed_dim,
                dim_head=embed_dim,
                heads=self.args.num_heads,
            )
            self.gru = PCBlock4_Deep_nopool_res(
                embed_dim * 4, embed_dim, k_conv=args.PCUpdater_conv
            )
        else:
            self.gru = PCBlock4_Deep_nopool_res(
                embed_dim * 4, embed_dim, k_conv=args.PCUpdater_conv
            )

        self.mask = nn.Sequential(
            nn.Conv2d(embed_dim, embed_dim * 2, 3, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(embed_dim * 2, ratio * ratio * 9, 1, padding=0),
        )

        self.temporal_attention1 = TemporalLayer(dim=embed_dim * 3)
        self.temporal_attention2 = TemporalLayer(dim=embed_dim)
        self.temporal_flow_head = TemporalFlowHead(embed_dim, 2, args.k_conv)

    def forward(self, nets, inps, corrs, flows, attentions, ref_nets=None, T=None):
        BT, _, H, W = nets.shape
        B = BT // T

        motion_features = self.encoder(flows, corrs, attentions)
        motion_features_globals = self.aggregator(attentions, motion_features)
        inp_cats = torch.cat([inps, motion_features, motion_features_globals], dim=1)

        # Temporal Attention
        inp_cats = rearrange(inp_cats, "(B T) C H W -> B T C H W", B=B, T=T)
        inp_cats = self.temporal_attention1(inp_cats)
        # Attentional update
        nets = self.gru(torch.cat([nets, inp_cats], dim=1))
        nets = rearrange(nets, "(B T) C H W -> B T C H W", B=B, T=T)
        nets = self.temporal_attention2(nets)

        nets = rearrange(nets, "(B T) C H W -> B T C H W", B=B, T=T)
        delta_flows = self.temporal_flow_head(nets)

        # scale mask to balence gradients
        nets = rearrange(nets, "B T C H W -> (B T) C H W")
        masks = 0.25 * self.mask(nets)

        masks = rearrange(masks, "(B T) C H W -> B T C H W", B=B, T=T)
        delta_flows = rearrange(delta_flows, "(B T) C H W -> B T C H W ", B=B, T=T)
        return nets, masks, delta_flows


class TemporalFusion(nn.Module):
    def __init__(self, args):
        super().__init__()
        self.fusion1 = args.fusion1
        self.proj = nn.Sequential(
            nn.Conv2d(
                args.decoder_dim // 2, args.decoder_dim // 2, kernel_size=3, padding=1
            ),
            nn.GELU(),
        )
        # init
        self.gamma = nn.Parameter(torch.zeros(1))

    def forward(self, x, temporal_refs):
        # x: [B, C, H, W]; temporal_refs: [B, T, C, H, W]; (x + temporal_refs) => (B, T, C, H, W) (B, C, H, W, T)
        if self.fusion1:
            ref = torch.mean(temporal_refs, dim=1, keepdim=False)
        else:
            res_x = x.unsqueeze(1).expand_as(temporal_refs)
            res_x = torch.abs(res_x - temporal_refs).sum(
                dim=(2, 3, 4)
            )  # 遮挡最严重的部分
            _, res_index = torch.max(res_x, dim=1)
            a = []
            for i in range(res_index.shape[0]):
                a.append(temporal_refs[i, res_index[i]])
            ref = torch.stack(a, dim=0)
            # print(ref.shape, res_index)
        x = x + self.gamma * self.proj(ref)

        return x


class SKFlowDecoder_MMBank(nn.Module):
    def __init__(self, args):
        super().__init__()
        self.args = args
        self.encoder = SKMotionEncoder6_Deep_nopool_res(args)
        ratio = 16 if args.Encoder == "UMT" else 8

        self.gma = args.use_gma
        embed_dim = args.decoder_dim // 2

        if self.gma:
            self.aggregator = Aggregate(
                args=self.args,
                dim=embed_dim,
                dim_head=embed_dim,
                heads=self.args.num_heads,
            )
            self.gru = PCBlock4_Deep_nopool_res(
                embed_dim * 4, embed_dim, k_conv=args.PCUpdater_conv
            )
        else:
            self.gru = PCBlock4_Deep_nopool_res(
                embed_dim * 4, embed_dim, k_conv=args.PCUpdater_conv
            )

        self.flow_head = PCBlock4_Deep_nopool_res(embed_dim, 2, args.k_conv)

        self.mask = nn.Sequential(
            nn.Conv2d(embed_dim, embed_dim * 2, 3, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(embed_dim * 2, ratio * ratio * 9, 1, padding=0),
        )

        # self.temporal_fusion = TemporalFusion(args) # TODO

    def forward(self, net, inp, corr, flow, attention, ref_nets):
        motion_features = self.encoder(flow, corr, attention)

        if self.gma:
            motion_features_global = self.aggregator(attention, motion_features)
            inp_cat = torch.cat([inp, motion_features, motion_features_global], dim=1)
        else:
            inp_cat = torch.cat([inp, motion_features], dim=1)
        # Attentional update

        # temporal Attention?
        # MemoryBank

        net = self.gru(torch.cat([net, inp_cat], dim=1))
        net = self.temporal_fusion(net, ref_nets)

        delta_flow = self.flow_head(net)

        # scale mask to balence gradients
        mask = 0.25 * self.mask(net)
        return net, mask, delta_flow


class SKFlowDecoder_planB(nn.Module):
    def __init__(self, args):
        super().__init__()
        self.args = args
        self.encoder = SKMotionEncoder6_Deep_nopool_res(args)
        ratio = 16 if args.Encoder == "UMT" else 8

        self.gma = args.use_gma
        embed_dim = args.decoder_dim // 2

        if self.gma:
            self.aggregator = Aggregate(
                args=self.args,
                dim=embed_dim,
                dim_head=embed_dim,
                heads=self.args.num_heads,
            )
            self.gru = PCBlock4_Deep_nopool_res(
                embed_dim * 4, embed_dim, k_conv=args.PCUpdater_conv
            )
        else:
            self.gru = PCBlock4_Deep_nopool_res(
                embed_dim * 4, embed_dim, k_conv=args.PCUpdater_conv
            )

        self.flow_head = PCBlock4_Deep_nopool_res(embed_dim, 2, args.k_conv)

        self.mask = nn.Sequential(
            nn.Conv2d(embed_dim, embed_dim * 2, 3, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(embed_dim * 2, ratio * ratio * 9, 1, padding=0),
        )

        self.temporal_fusion = TemporalFusion(args)

    def forward(self, net, inp, corr, flow, attention, ref_nets):
        motion_features = self.encoder(flow, corr, attention)

        if self.gma:
            motion_features_global = self.aggregator(attention, motion_features)
            inp_cat = torch.cat([inp, motion_features, motion_features_global], dim=1)
        else:
            inp_cat = torch.cat([inp, motion_features], dim=1)
        # Attentional update
        net = self.gru(torch.cat([net, inp_cat], dim=1))
        net = self.temporal_fusion(net, ref_nets)

        delta_flow = self.flow_head(net)

        # scale mask to balence gradients
        mask = 0.25 * self.mask(net)
        return net, mask, delta_flow


class FlowHead(nn.Module):
    def __init__(self, input_dim=128, hidden_dim=256):
        super(FlowHead, self).__init__()
        self.conv1 = nn.Conv2d(input_dim, hidden_dim, 3, padding=1)
        self.conv2 = nn.Conv2d(hidden_dim, 2, 3, padding=1)
        self.relu = nn.ReLU(inplace=True)

    def forward(self, x):
        return self.conv2(self.relu(self.conv1(x)))


class FlowHeadNew(nn.Module):
    def __init__(self, C_in, C_out):
        super().__init__()
        hidden_dim = 128
        self.conv1 = nn.Conv2d(C_in, hidden_dim, 3, padding=1)
        self.conv2 = nn.Conv2d(hidden_dim, C_out, 3, padding=1)
        self.relu = nn.ReLU(inplace=True)

    def forward(self, x):
        return self.conv2(self.relu(self.conv1(x)))


class SepConvGRU(nn.Module):
    def __init__(self, hidden_dim=128, input_dim=192 + 128):
        super(SepConvGRU, self).__init__()
        self.convz1 = nn.Conv2d(
            hidden_dim + input_dim, hidden_dim, (1, 5), padding=(0, 2)
        )
        self.convr1 = nn.Conv2d(
            hidden_dim + input_dim, hidden_dim, (1, 5), padding=(0, 2)
        )
        self.convq1 = nn.Conv2d(
            hidden_dim + input_dim, hidden_dim, (1, 5), padding=(0, 2)
        )

        self.convz2 = nn.Conv2d(
            hidden_dim + input_dim, hidden_dim, (5, 1), padding=(2, 0)
        )
        self.convr2 = nn.Conv2d(
            hidden_dim + input_dim, hidden_dim, (5, 1), padding=(2, 0)
        )
        self.convq2 = nn.Conv2d(
            hidden_dim + input_dim, hidden_dim, (5, 1), padding=(2, 0)
        )

    def forward(self, h, x):
        # horizontal
        hx = torch.cat([h, x], dim=1)
        z = torch.sigmoid(self.convz1(hx))
        r = torch.sigmoid(self.convr1(hx))
        q = torch.tanh(self.convq1(torch.cat([r * h, x], dim=1)))
        h = (1 - z) * h + z * q

        # vertical
        hx = torch.cat([h, x], dim=1)
        z = torch.sigmoid(self.convz2(hx))
        r = torch.sigmoid(self.convr2(hx))
        q = torch.tanh(self.convq2(torch.cat([r * h, x], dim=1)))
        h = (1 - z) * h + z * q

        return h


class BasicMotionEncoder(nn.Module):
    def __init__(self, args):
        super(BasicMotionEncoder, self).__init__()
        cor_planes = args.corr_levels * (2 * args.corr_radius + 1) ** 2
        if "All" in args.model_name or args.cost_encoder_v1:
            cor_planes += 128
        elif "cost_encoder" in args.model_name or args.cost_encoder_v2:
            cor_planes = 512

        self.convc1 = nn.Conv2d(cor_planes, 256, 1, padding=0)
        self.convc2 = nn.Conv2d(256, 192, 3, padding=1)
        self.convf1 = nn.Conv2d(2, 128, 7, padding=3)
        self.convf2 = nn.Conv2d(128, 64, 3, padding=1)
        self.conv = nn.Conv2d(64 + 192, 128 - 2, 3, padding=1)

    def forward(self, flow, corr):
        cor = F.relu(self.convc1(corr))
        cor = F.relu(self.convc2(cor))
        flo = F.relu(self.convf1(flow))
        flo = F.relu(self.convf2(flo))

        cor_flo = torch.cat([cor, flo], dim=1)
        out = F.relu(self.conv(cor_flo))
        return torch.cat([out, flow], dim=1)


class RAFTDeepMotionEncoder(nn.Module):
    def __init__(self, args):
        super().__init__()
        cor_planes = args.corr_levels * (2 * args.corr_radius + 1) ** 2
        if "All" in args.model_name or args.cost_encoder_v1:
            cor_planes += 128
        elif "cost_encoder" in args.model_name or args.cost_encoder_v2:
            cor_planes = 512

        self.convc1 = nn.Conv2d(cor_planes, 256, 1, padding=0)
        self.convc2 = nn.Conv2d(256, 256, 3, padding=1)
        self.convc3 = nn.Conv2d(256, 192, 3, padding=1)
        self.convf1 = nn.Conv2d(2, 128, 7, padding=3)
        self.convf2 = nn.Conv2d(128, 128, 3, padding=1)
        self.convf3 = nn.Conv2d(128, 64, 3, padding=1)
        self.conv = nn.Conv2d(64 + 192, 144, 3, padding=1)
        # self.conv2 = nn.Conv2d(192, 192, 3, padding=1)
        self.conv2 = nn.Conv2d(144, 128 - 2, 3, padding=1)
        # self.conv3 = nn.Conv2d(126, 126, 3, padding=1)

    def forward(self, flow, corr):
        cor = F.relu(self.convc1(corr))
        cor = F.relu(self.convc2(cor))
        cor = F.relu(self.convc3(cor))
        flo = F.relu(self.convf1(flow))
        flo = F.relu(self.convf2(flo))
        flo = F.relu(self.convf3(flo))

        cor_flo = torch.cat([cor, flo], dim=1)
        out = F.relu(self.conv(cor_flo))
        out = F.relu(self.conv2(out))
        # out = F.relu(self.conv3(out))
        return torch.cat([out, flow], dim=1)


class DeepMotionEncoder(nn.Module):
    def __init__(self, args):
        super().__init__()
        cor_planes = args.corr_levels * (2 * args.corr_radius + 1) ** 2
        if "All" in args.model_name or args.cost_encoder_v1:
            cor_planes += 128
        elif "cost_encoder" in args.model_name or args.cost_encoder_v2:
            cor_planes = 512

        self.convc1 = nn.Conv2d(cor_planes, 256, 1, padding=0)
        self.convc2 = nn.Conv2d(256, 256, 3, padding=1)
        self.convc3 = nn.Conv2d(256, 192, 3, padding=1)
        self.convf1 = nn.Conv2d(2, 128, 7, padding=3)
        self.convf2 = nn.Conv2d(128, 128, 3, padding=1)
        self.convf3 = nn.Conv2d(128, 64, 3, padding=1)
        self.conv = nn.Conv2d(64 + 192, 192, 3, padding=1)
        self.conv2 = nn.Conv2d(192, 192, 3, padding=1)
        self.conv3 = nn.Conv2d(192, 128 - 2, 3, padding=1)
        # self.conv3 = nn.Conv2d(126, 126, 3, padding=1)

    def forward(self, flow, corr):
        cor = F.relu(self.convc1(corr))
        cor = F.relu(self.convc2(cor))
        cor = F.relu(self.convc3(cor))
        flo = F.relu(self.convf1(flow))
        flo = F.relu(self.convf2(flo))
        flo = F.relu(self.convf3(flo))

        cor_flo = torch.cat([cor, flo], dim=1)
        out = F.relu(self.conv(cor_flo))
        out = F.relu(self.conv2(out))
        out = F.relu(self.conv3(out))
        return torch.cat([out, flow], dim=1)


class BasicUpdateBlock(nn.Module):
    def __init__(self, args, hidden_dim=128, input_dim=128):
        super(BasicUpdateBlock, self).__init__()
        self.args = args
        self.encoder = BasicMotionEncoder(args)
        self.gru = SepConvGRU(hidden_dim=hidden_dim, input_dim=128 + hidden_dim)
        self.flow_head = FlowHead(hidden_dim, hidden_dim=256)

        self.mask = nn.Sequential(
            nn.Conv2d(128, 256, 3, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(256, 64 * 9, 1, padding=0),
        )

    def forward(self, net, inp, corr, flow, upsample=True):
        motion_features = self.encoder(flow, corr)
        inp = torch.cat([inp, motion_features], dim=1)

        net = self.gru(net, inp)
        delta_flow = self.flow_head(net)

        # scale mask to balence gradients
        mask = 0.25 * self.mask(net)
        return net, mask, delta_flow


class RAFTDeepUpdateBlock(nn.Module):
    def __init__(self, args, hidden_dim=128, input_dim=128):
        super().__init__()
        self.args = args
        self.encoder = RAFTDeepMotionEncoder(args)
        self.gru = SepConvGRU(hidden_dim=hidden_dim, input_dim=128 + hidden_dim)
        self.flow_head = FlowHead(hidden_dim, hidden_dim=256)

        self.mask = nn.Sequential(
            nn.Conv2d(128, 256, 3, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(256, 64 * 9, 1, padding=0),
        )

    def forward(self, net, inp, corr, flow, upsample=True):
        motion_features = self.encoder(flow, corr)
        inp = torch.cat([inp, motion_features], dim=1)

        net = self.gru(net, inp)
        delta_flow = self.flow_head(net)

        # scale mask to balence gradients
        mask = 0.25 * self.mask(net)
        return net, mask, delta_flow


class GMADeepUpdateBlock(nn.Module):
    def __init__(self, args, hidden_dim=128, input_dim=128):
        super().__init__()
        self.args = args
        self.encoder = DeepMotionEncoder(args)
        self.gru = SepConvGRU(hidden_dim=hidden_dim, input_dim=128 + hidden_dim)
        self.flow_head = FlowHead(hidden_dim, hidden_dim=256)

        self.mask = nn.Sequential(
            nn.Conv2d(128, 256, 3, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(256, 64 * 9, 1, padding=0),
        )

    def forward(self, net, inp, corr, flow, upsample=True):
        motion_features = self.encoder(flow, corr)
        inp = torch.cat([inp, motion_features], dim=1)

        net = self.gru(net, inp)
        delta_flow = self.flow_head(net)

        # scale mask to balence gradients
        mask = 0.25 * self.mask(net)
        return net, mask, delta_flow


# CBAM
class ChannelAttention(nn.Module):
    def __init__(self, in_planes, ratio=16):
        super(ChannelAttention, self).__init__()
        self.avg_pool = nn.AdaptiveAvgPool2d(1)
        self.max_pool = nn.AdaptiveMaxPool2d(1)

        self.fc = nn.Sequential(
            nn.Conv2d(in_planes, in_planes // 16, 1, bias=False),
            nn.ReLU(),
            nn.Conv2d(in_planes // 16, in_planes, 1, bias=False),
        )
        self.sigmoid = nn.Sigmoid()

    def forward(self, x):
        avg_out = self.fc(self.avg_pool(x))
        max_out = self.fc(self.max_pool(x))
        out = avg_out + max_out
        return self.sigmoid(out)


class SpatialAttention(nn.Module):
    def __init__(self, kernel_size=7):
        super(SpatialAttention, self).__init__()

        self.conv1 = nn.Conv2d(2, 1, kernel_size, padding=kernel_size // 2, bias=False)
        self.sigmoid = nn.Sigmoid()

    def forward(self, x):
        avg_out = torch.mean(x, dim=1, keepdim=True)
        max_out, _ = torch.max(x, dim=1, keepdim=True)
        x = torch.cat([avg_out, max_out], dim=1)
        x = self.conv1(x)
        return self.sigmoid(x)


class SKBlock_CBAM(nn.Module):
    def __init__(self, C_in, C_out, k_conv):
        super().__init__()
        self.conv_list = nn.ModuleList(
            [
                nn.Conv2d(
                    C_in, C_in, kernel, stride=1, padding=kernel // 2, groups=C_in
                )
                for kernel in k_conv
            ]
        )

        self.ffn1 = nn.Sequential(
            nn.Conv2d(C_in, int(1.5 * C_in), 1, padding=0),
            nn.GELU(),
            nn.Conv2d(int(1.5 * C_in), C_in, 1, padding=0),
        )
        self.pw = nn.Conv2d(C_in, C_in, 1, padding=0)
        self.ffn2 = nn.Sequential(
            nn.Conv2d(C_in, int(1.5 * C_in), 1, padding=0),
            nn.GELU(),
            nn.Conv2d(int(1.5 * C_in), C_out, 1, padding=0),
        )
        self.ca = ChannelAttention(C_in)
        self.sa = SpatialAttention()

    def forward(self, x):
        x = F.gelu(x + self.ffn1(x))
        for conv in self.conv_list:
            x = F.gelu(x + conv(x))
        x = F.gelu(x + self.pw(x))

        # CBAM
        in_feature = x
        x = self.ca(x) * x
        x = self.sa(x) * x
        x = F.gelu(x + in_feature)

        x = self.ffn2(x)
        return x


class SKMotionEncoder_CBAM(nn.Module):
    def __init__(self, args):
        super().__init__()
        cor_planes = args.corr_levels * (2 * args.corr_radius + 1) ** 2
        self.convc1 = SKBlock_CBAM(cor_planes, 256, args.k_conv)
        self.convc2 = SKBlock_CBAM(256, 192, args.k_conv)

        self.convf1 = nn.Conv2d(2, 128, 1, 1, 0)
        self.convf2 = SKBlock_CBAM(128, 64, args.k_conv)

        self.conv = SKBlock_CBAM(64 + 192, 128 - 2, args.k_conv)

    def forward(self, flow, corr):
        cor = F.gelu(self.convc1(corr))

        cor = self.convc2(cor)

        flo = self.convf1(flow)
        flo = self.convf2(flo)

        cor_flo = torch.cat([cor, flo], dim=1)
        out = self.conv(cor_flo)

        return torch.cat([out, flow], dim=1)


class SKUpdateBlock_CBAM(nn.Module):
    def __init__(self, args, hidden_dim):
        super().__init__()
        self.args = args
        self.encoder = SKMotionEncoder_CBAM(args)

        self.gru = SKBlock_CBAM(
            128 + hidden_dim + hidden_dim + 128, 128, k_conv=args.PCUpdater_conv
        )
        self.flow_head = SKBlock_CBAM(128, 2, args.k_conv)

        self.mask = nn.Sequential(
            nn.Conv2d(128, 256, 3, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(256, 64 * 9, 1, padding=0),
        )

        self.aggregator = Aggregate(
            args=self.args, dim=128, dim_head=128, heads=self.args.num_heads
        )

    def forward(self, net, inp, corr, flow, attention):
        motion_features = self.encoder(flow, corr)
        motion_features_global = self.aggregator(attention, motion_features)
        inp_cat = torch.cat([inp, motion_features, motion_features_global], dim=1)

        # Attentional update
        net = self.gru(torch.cat([net, inp_cat], dim=1))

        delta_flow = self.flow_head(net)

        # scale mask to balence gradients
        mask = 0.25 * self.mask(net)
        return net, mask, delta_flow


# temporal flow head v1
class GMAUpdateBlock_TAM_v7(nn.Module):
    def __init__(self, args):
        super().__init__()
        self.args = args
        self.encoder = BasicMotionEncoder(args)
        ratio = 16 if args.Encoder == "UMT" else 8

        self.gma = args.use_gma
        embed_dim = args.decoder_dim // 2
        hidden_dim = embed_dim

        if self.gma:
            self.aggregator = Aggregate(
                args=self.args,
                dim=embed_dim,
                dim_head=embed_dim,
                heads=self.args.num_heads,
            )
            self.gru = SepConvGRU(
                hidden_dim=hidden_dim,
                input_dim=128 + hidden_dim + hidden_dim + hidden_dim,
            )
        else:
            self.gru = SepConvGRU(
                hidden_dim=hidden_dim, input_dim=128 + hidden_dim + hidden_dim
            )

        self.mask = nn.Sequential(
            nn.Conv2d(embed_dim, embed_dim * 2, 3, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(embed_dim * 2, ratio * ratio * 9, 1, padding=0),
        )

        self.transformer_block = TemporalLayer2(dim=embed_dim)
        self.temporal_regressor = TemporalLayer2(dim=2)
        self.flow_head = FlowHeadNew(embed_dim * (args.T - 1), 2 * (args.T - 1))

    def forward(self, nets, inps, corrs, flows, attentions, T=None):
        BT, _, H, W = nets.shape
        B = BT // T

        motion_features = self.encoder(flows, corrs)
        motion_features_temporal = self.transformer_block(
            rearrange(motion_features, "(B T) C H W -> (B H W) T C", T=T), HW=(H, W)
        )

        if self.gma:
            motion_features_globals = self.aggregator(attentions, motion_features)
            inp_cats = torch.cat(
                [
                    inps,
                    motion_features,
                    motion_features_globals,
                    motion_features_temporal,
                ],
                dim=1,
            )
        else:
            inp_cats = torch.cat(
                [inps, motion_features, motion_features_temporal], dim=1
            )
        # Temporal Attention
        # Attentional update
        nets = self.gru(nets, inp_cats)
        delta_flows = self.flow_head(
            rearrange(nets, "(B T) C H W -> B (T C) H W", B=B)
        )  # (b t) c h w => b (t c) h w

        # scale mask to balence gradients
        masks = 0.25 * self.mask(nets)

        masks = rearrange(masks, "(B T) C H W -> B T C H W", B=B, T=T)
        delta_flows = rearrange(delta_flows, "B (T C) H W -> (B H W) T C ", T=T)
        delta_flows = self.temporal_regressor(delta_flows, HW=(H, W))
        delta_flows = rearrange(delta_flows, "(B T) C H W -> B T C H W", B=B)

        return nets, masks, delta_flows


# temporal flow head v1
class GMAUpdateBlock_TAM_ablation(nn.Module):
    def __init__(self, args):
        super().__init__()
        self.args = args
        self.encoder = BasicMotionEncoder(args)
        ratio = 16 if args.Encoder == "UMT" else 8

        self.gma = args.use_gma
        embed_dim = args.decoder_dim // 2
        hidden_dim = embed_dim

        if self.gma:
            self.aggregator = Aggregate(
                args=self.args,
                dim=embed_dim,
                dim_head=embed_dim,
                heads=self.args.num_heads,
            )
            self.gru = SepConvGRU(
                hidden_dim=hidden_dim,
                input_dim=128 + hidden_dim + hidden_dim + hidden_dim,
            )
        else:
            self.gru = SepConvGRU(
                hidden_dim=hidden_dim, input_dim=128 + hidden_dim + hidden_dim
            )

        self.mask = nn.Sequential(
            nn.Conv2d(embed_dim, embed_dim * 2, 3, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(embed_dim * 2, ratio * ratio * 9, 1, padding=0),
        )

        self.transformer_block = TemporalLayer_Ablation(dim=embed_dim)
        self.flow_head = FlowHeadNew(embed_dim * (args.T - 1), 2 * (args.T - 1))

    def forward(self, nets, inps, corrs, flows, attentions, T=None):
        BT, _, H, W = nets.shape
        B = BT // T

        motion_features = self.encoder(flows, corrs)
        motion_features_temporal = self.transformer_block(
            rearrange(motion_features, "(B T) C H W -> (B T) (H W) C", T=T), HW=(H, W)
        )

        if self.gma:
            motion_features_globals = self.aggregator(attentions, motion_features)
            inp_cats = torch.cat(
                [
                    inps,
                    motion_features,
                    motion_features_globals,
                    motion_features_temporal,
                ],
                dim=1,
            )
        else:
            inp_cats = torch.cat(
                [inps, motion_features, motion_features_temporal], dim=1
            )
        # Temporal Attention
        # Attentional update
        nets = self.gru(nets, inp_cats)
        delta_flows = self.flow_head(
            rearrange(nets, "(B T) C H W -> B (T C) H W", B=B)
        )  # (b t) c h w => b (t c) h w

        # scale mask to balence gradients
        masks = 0.25 * self.mask(nets)

        masks = rearrange(masks, "(B T) C H W -> B T C H W", B=B, T=T)
        delta_flows = rearrange(delta_flows, "B (T C) H W -> B T C H W", T=T)

        return nets, masks, delta_flows


class GMAUpdateBlock(nn.Module):
    def __init__(self, args, hidden_dim=128):
        super().__init__()
        self.args = args
        self.encoder = BasicMotionEncoder(args)
        self.gru = SepConvGRU(
            hidden_dim=hidden_dim, input_dim=128 + hidden_dim + hidden_dim
        )
        self.flow_head = FlowHead(hidden_dim, hidden_dim=256)

        self.mask = nn.Sequential(
            nn.Conv2d(128, 256, 3, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(256, 64 * 9, 1, padding=0),
        )

        self.aggregator = Aggregate(
            args=self.args, dim=128, dim_head=128, heads=self.args.num_heads
        )

    def forward(self, net, inp, corr, flow, attention, T=None):
        motion_features = self.encoder(flow, corr)
        motion_features_global = self.aggregator(attention, motion_features)
        inp_cat = torch.cat([inp, motion_features, motion_features_global], dim=1)

        # Attentional update
        net = self.gru(net, inp_cat)

        # corr关注的是非遮挡区域，所以和motion feature极其相似?。。。。。。。。。。。。。
        delta_flow = self.flow_head(net)

        # plot
        # plot_featmap(motion_features.cpu().squeeze(0), '/mnt/cloud_disk/ssk/playground/GMA-main/nips-figs/RAFTGMA/487-motion_feature')
        # plot_featmap(motion_features_global.cpu().squeeze(0), '/mnt/cloud_disk/ssk/playground/GMA-main/nips-figs/RAFTGMA/487-motion_feature_global')
        # plot_featmap(net.cpu().squeeze(0), '/mnt/cloud_disk/ssk/playground/GMA-main/nips-figs/RAFTGMA/487-net')

        # scale mask to balence gradients
        mask = 0.25 * self.mask(net)
        return net, mask, delta_flow


class RAFTUpdateBlock(nn.Module):
    def __init__(self, args, hidden_dim=128):
        super().__init__()
        self.args = args
        self.encoder = BasicMotionEncoder(args)
        self.gru = SepConvGRU(hidden_dim=hidden_dim, input_dim=128 + hidden_dim)
        self.flow_head = FlowHead(hidden_dim, hidden_dim=256)

        self.mask = nn.Sequential(
            nn.Conv2d(128, 256, 3, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(256, 64 * 9, 1, padding=0),
        )

        self.aggregator = Aggregate(
            args=self.args, dim=128, dim_head=128, heads=self.args.num_heads
        )

    def forward(self, net, inp, corr, flow, attention, T=None):
        motion_features = self.encoder(flow, corr)
        inp_cat = torch.cat([inp, motion_features], dim=1)

        # Attentional update
        net = self.gru(net, inp_cat)

        # corr关注的是非遮挡区域，所以和motion feature极其相似?。。。。。。。。。。。。。
        delta_flow = self.flow_head(net)

        # plot
        # plot_featmap(motion_features.cpu().squeeze(0), '/mnt/cloud_disk/ssk/playground/GMA-main/nips-figs/RAFTGMA/487-motion_feature')
        # plot_featmap(motion_features_global.cpu().squeeze(0), '/mnt/cloud_disk/ssk/playground/GMA-main/nips-figs/RAFTGMA/487-motion_feature_global')
        # plot_featmap(net.cpu().squeeze(0), '/mnt/cloud_disk/ssk/playground/GMA-main/nips-figs/RAFTGMA/487-net')

        # scale mask to balence gradients
        mask = 0.25 * self.mask(net)
        return net, mask, delta_flow


class MFUpdateBlock(nn.Module):
    def __init__(self, args, hidden_dim):
        super().__init__()
        self.args = args
        self.gru = SepConvGRU(hidden_dim=hidden_dim, input_dim=128 + hidden_dim)
        self.flow_head = FlowHead(hidden_dim, hidden_dim=256)

        self.mask = nn.Sequential(
            nn.Conv2d(128, 256, 3, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(256, 64 * 9, 1, padding=0),
        )

    # corr: 第i和i+1; corr_before: 第i-1和i
    def forward(self, net, inp, motion_features, attention=None):
        inp_cat = torch.cat([inp, motion_features], dim=1)

        # Attentional update
        net = self.gru(net, inp_cat)

        delta_flow = self.flow_head(net)

        # scale mask to balence gradients
        mask = 0.25 * self.mask(net)
        return net, mask, delta_flow


class SKFlowMotionEncoder(nn.Module):
    def __init__(self, args):
        super().__init__()
        self.args = args
        cor_planes = args.corr_levels * (2 * args.corr_radius + 1) ** 2

        self.convc1 = PCBlock4_Deep_nopool_res(cor_planes, 256, args.k_conv)
        self.convc2 = PCBlock4_Deep_nopool_res(256, 192, args.k_conv)

        self.convf1 = nn.Conv2d(2, 128, 1, 1, 0)
        self.convf2 = PCBlock4_Deep_nopool_res(128, 64, args.k_conv)

        self.conv = PCBlock4_Deep_nopool_res(64 + 192, 128 - 2, args.k_conv)

    def forward(self, flow, corr, attention=None):
        cor = F.gelu(self.convc1(corr))

        cor = self.convc2(cor)

        flo = self.convf1(flow)
        flo = self.convf2(flo)

        cor_flo = torch.cat([cor, flo], dim=1)
        out = self.conv(cor_flo)

        return torch.cat([out, flow], dim=1)


class MFSKFlowUpdateBlock(nn.Module):
    def __init__(self, args, hidden_dim):
        super().__init__()
        self.args = args
        if args.use_temporal_decoder:
            self.aggregator = SpatioTemporalAggregate(
                args=self.args, dim=128, dim_head=128, heads=self.args.num_heads
            )
        else:
            self.aggregator = Aggregate(
                args=self.args, dim=128, dim_head=128, heads=self.args.num_heads
            )
        self.gru = PCBlock4_Deep_nopool_res(
            128 + hidden_dim + hidden_dim + 128, 128, k_conv=args.PCUpdater_conv
        )
        self.flow_head = PCBlock4_Deep_nopool_res(128, 2, args.k_conv)

        self.mask = nn.Sequential(
            nn.Conv2d(128, 256, 3, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(256, 64 * 9, 1, padding=0),
        )

    def forward(self, net, inp, motion_features, attention, temporal_attention=None):
        if self.args.use_temporal_decoder:
            motion_features_global = self.aggregator(
                attention, temporal_attention, motion_features
            )
        else:
            motion_features_global = self.aggregator(attention, motion_features)

        inp_cat = torch.cat([inp, motion_features, motion_features_global], dim=1)

        # Attentional update
        net = self.gru(torch.cat([net, inp_cat], dim=1))

        delta_flow = self.flow_head(net)

        # scale mask to balence gradients
        mask = 0.25 * self.mask(net)
        return net, mask, delta_flow


class MFRAFTUpdateBlock(nn.Module):
    def __init__(self, args, hidden_dim):
        super().__init__()
        self.args = args
        if args.use_temporal_decoder:
            self.aggregator = TemporalAggregate(
                args=self.args, dim=128, dim_head=128, heads=self.args.num_heads
            )
            if self.args.fuse2:
                self.gru = SepConvGRU(hidden_dim=hidden_dim, input_dim=128 + hidden_dim)
            else:
                self.temporal_gru = SepConvGRU(
                    hidden_dim=hidden_dim, input_dim=128 + hidden_dim + hidden_dim
                )
        else:
            self.aggregator = None
            self.gru = SepConvGRU(hidden_dim=hidden_dim, input_dim=128 + hidden_dim)
        self.flow_head = FlowHead(hidden_dim, hidden_dim=256)

        self.mask = nn.Sequential(
            nn.Conv2d(128, 256, 3, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(256, 64 * 9, 1, padding=0),
        )

    def forward(self, net, inp, motion_features, temporal_attention=None):
        if self.args.use_temporal_decoder:
            motion_features_global = self.aggregator(
                temporal_attention, motion_features
            )
            if self.args.fuse2:
                motion_features = motion_features + motion_features_global
                inp_cat = torch.cat([inp, motion_features], dim=1)
                net = self.gru(net, inp_cat)
            else:
                inp_cat = torch.cat(
                    [inp, motion_features, motion_features_global], dim=1
                )
                net = self.temporal_gru(net, inp_cat)
        else:
            inp_cat = torch.cat([inp, motion_features], dim=1)
            net = self.gru(net, inp_cat)
        # Attentional update
        delta_flow = self.flow_head(net)

        # scale mask to balence gradients
        mask = 0.25 * self.mask(net)
        return net, mask, delta_flow


class MFGMAUpdateBlock(nn.Module):
    def __init__(self, args, hidden_dim):
        super().__init__()
        self.args = args
        if args.use_temporal_decoder:
            self.aggregator = SpatioTemporalAggregate(
                args=self.args, dim=128, dim_head=128, heads=self.args.num_heads
            )
        else:
            self.aggregator = Aggregate(
                args=self.args, dim=128, dim_head=128, heads=self.args.num_heads
            )
        self.gru = SepConvGRU(
            hidden_dim=hidden_dim, input_dim=128 + hidden_dim + hidden_dim
        )
        self.flow_head = FlowHead(hidden_dim, hidden_dim=256)

        self.mask = nn.Sequential(
            nn.Conv2d(128, 256, 3, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(256, 64 * 9, 1, padding=0),
        )

    def forward(self, net, inp, motion_features, attention, temporal_attention=None):
        if self.args.use_temporal_decoder:
            motion_features_global = self.aggregator(
                attention, temporal_attention, motion_features
            )
        else:
            motion_features_global = self.aggregator(attention, motion_features)

        # Attentional update
        inp_cat = torch.cat([inp, motion_features, motion_features_global], dim=1)
        net = self.gru(net, inp_cat)

        delta_flow = self.flow_head(net)

        # scale mask to balence gradients
        mask = 0.25 * self.mask(net)
        return net, mask, delta_flow


class ControlMFSKIIUpdateBlock(nn.Module):
    def __init__(self, args, hidden_dim):
        super().__init__()
        self.args = args
        self.aggregator = SpatioTemporalAggregate(
            args=self.args, dim=128, dim_head=128, heads=self.args.num_heads
        )
        self.gru = eval(args.SKII_Block)(
            128 + hidden_dim + hidden_dim + 128, 128, args.PCUpdater_conv, args
        )
        self.flow_head = eval(args.SKII_Block)(128, 2, args.k_conv, args)

        self.mask = nn.Sequential(
            nn.Conv2d(128, 256, 3, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(256, 64 * 9, 1, padding=0),
        )

        self.zero_conv0 = nn.Conv2d(128, 128, 1)
        self.zero_conv1 = nn.Conv2d(2, 2, 1)
        self.zero_conv2 = nn.Conv2d(64 * 9, 64 * 9, 1)

        for p in self.zero_conv0.parameters():
            nn.init.zeros_(p)
        for p in self.zero_conv1.parameters():
            nn.init.zeros_(p)
        for p in self.zero_conv2.parameters():
            nn.init.zeros_(p)

    def forward(self, net, inp, motion_features, attention, temporal_attention=None):
        motion_features_global = self.aggregator(
            attention, temporal_attention, motion_features
        )
        inp_cat = torch.cat([inp, motion_features, motion_features_global], dim=1)

        # Attentional update
        net = self.gru(torch.cat([net, inp_cat], dim=1))
        net = self.zero_conv0(net)

        delta_flow = self.flow_head(net)
        delta_flow = self.zero_conv1(delta_flow)

        # scale mask to balence gradients
        mask = 0.25 * self.mask(net)
        mask = self.zero_conv2(mask)
        return net, mask, delta_flow



# =============================================================================
# Copyright 2024 Henrique Morimitsu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Adapted from RAFT: https://github.com/princeton-vl/RAFT
#
# Original license BSD 3-clause: https://github.com/princeton-vl/RAFT/blob/master/LICENSE
#
# Modifications by Henrique Morimitsu:
# - Adapt model parameters
# =============================================================================

import torch
import torch.nn as nn
import torch.nn.functional as F

from .next1d import NeXt1DStage
from .local_timm.norm import LayerNorm2d


class FlowHead(nn.Module):
    def __init__(self, input_dim=128, hidden_dim=256):
        super(FlowHead, self).__init__()
        self.conv1 = nn.Conv2d(input_dim, hidden_dim, 3, padding=1)
        self.conv2 = nn.Conv2d(hidden_dim, 2, 3, padding=1)
        self.relu = nn.ReLU(inplace=True)

    def forward(self, x):
        return self.conv2(self.relu(self.conv1(x)))


class NeXT1DDecoder(nn.Module):
    def __init__(
        self,
        hidden_dim=128,
        input_dim=192 + 128,
        ksize=7,
        depth=1,
        mlp_ratio=4.0,
        norm_layer=None,
        fuse_next1d_weights=False,
    ):
        super(NeXT1DDecoder, self).__init__()
        self.conv = NeXt1DStage(
            hidden_dim + input_dim,
            hidden_dim,
            kernel_size=ksize,
            stride=1,
            depth=depth,
            mlp_ratio=mlp_ratio,
            norm_layer=norm_layer,
            fuse_next1d_weights=fuse_next1d_weights,
        )

    def forward(self, h, x):
        hx = torch.cat([h, x], dim=1)
        h = self.conv(hx)
        h = torch.tanh(h)
        return h


class MotionEncoder(nn.Module):
    def __init__(self, corr_levels: int, corr_range: int, dec_motion_chs: int):
        super(MotionEncoder, self).__init__()

        c_hidden = 256
        c_out = 192
        f_hidden = 128
        f_out = 64

        cor_planes = corr_levels * (2 * corr_range + 1) ** 2
        self.convc1 = nn.Conv2d(cor_planes, c_hidden, 1, padding=0)
        self.convc2 = nn.Conv2d(c_hidden, c_out, 3, padding=1)
        self.convf1 = nn.Conv2d(2, f_hidden, 7, padding=3)
        self.convf2 = nn.Conv2d(f_hidden, f_out, 3, padding=1)
        self.conv = nn.Conv2d(f_out + c_out, dec_motion_chs - 2, 3, padding=1)

    def forward(self, flow, corr):
        cor = F.relu(self.convc1(corr))
        cor = F.relu(self.convc2(cor))
        flo = F.relu(self.convf1(flow))
        flo = F.relu(self.convf2(flo))

        cor_flo = torch.cat([cor, flo], dim=1)
        out = F.relu(self.conv(cor_flo))
        return torch.cat([out, flow], dim=1)


class UpdateBlock(nn.Module):
    def __init__(
        self,
        pyramid_ranges: tuple[int, int],
        corr_levels: int,
        corr_range: int,
        dec_net_chs: int,
        dec_inp_chs: int,
        dec_motion_chs: int,
        dec_depth: int,
        dec_mlp_ratio: float,
        fuse_next1d_weights: bool,
        use_upsample_mask: bool,
    ):
        super(UpdateBlock, self).__init__()
        self.use_upsample_mask = use_upsample_mask

        self.encoder = MotionEncoder(
            corr_levels=corr_levels,
            corr_range=corr_range,
            dec_motion_chs=dec_motion_chs,
        )
        self.decoder = NeXT1DDecoder(
            hidden_dim=dec_net_chs,
            input_dim=dec_motion_chs + dec_inp_chs,
            ksize=7,
            depth=dec_depth,
            mlp_ratio=dec_mlp_ratio,
            norm_layer=LayerNorm2d,
            fuse_next1d_weights=fuse_next1d_weights,
        )

        self.flow_head = FlowHead(dec_net_chs, hidden_dim=256)

        pred_stride = min(pyramid_ranges) if use_upsample_mask else 8
        self.mask = nn.Sequential(
            nn.Conv2d(dec_net_chs, dec_net_chs * 2, 3, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(dec_net_chs * 2, pred_stride**2 * 9, 1, padding=0),
        )

    def forward(self, net, inp, corr, flow, get_mask=False):
        motion_features = self.encoder(flow, corr)
        inp = torch.cat([inp, motion_features], dim=1)

        net = self.decoder(net, inp)
        delta_flow = self.flow_head(net)

        mask = None
        if self.use_upsample_mask and get_mask:
            mask = self.mask(net)

        return delta_flow, net, mask



# Copyright The Lightning AI team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# THIS FILE MUST READ EASILY, FOR UNDERSTANDING AND DEBUGGING PURPOSES.
# DO NOT OBSCURE THE TRAINING LOOP
# THIS IS A HARD REQUIREMENT TO CONTRIBUTING TO LIGHTNING
# WE FAVOR READABILITY OVER ENGINEERING-CONSTRUCTS BY DESIGN
# DO NOT REMOVE THIS NOTICE
# - WILLIAM FALCON
"""Trainer to automate the training."""

import logging
from datetime import timedelta
from typing import Dict, Iterable, List, Optional, Union
from weakref import proxy

import lightning.pytorch as pl
from lightning.fabric.utilities.types import _PATH
from lightning.pytorch.accelerators import Accelerator
from lightning.pytorch.callbacks import Callback
from lightning.pytorch.loggers import Logger
from lightning.pytorch.loggers.utilities import _log_hyperparams
from lightning.pytorch.loops.utilities import _parse_loop_limits
from lightning.pytorch.plugins import _PLUGIN_INPUT
from lightning.pytorch.profilers import Profiler
from lightning.pytorch.strategies import Strategy
from lightning.pytorch.trainer import call
from lightning.pytorch.trainer.configuration_validator import (
    _verify_loop_configurations,
)
from lightning.pytorch.trainer.connectors.accelerator_connector import (
    _LITERAL_WARN,
    _PRECISION_INPUT,
)
from lightning.pytorch.trainer.states import (
    TrainerFn,
    TrainerStatus,
)
from lightning.pytorch.utilities import parsing
from lightning.pytorch.utilities.argparse import _defaults_from_env_vars
from lightning.pytorch.utilities.rank_zero import rank_zero_info
from lightning.pytorch.utilities.types import (
    _EVALUATE_OUTPUT,
    _PREDICT_OUTPUT,
)

from .ptlflow_checkpoint_connector import (
    _PTLFlowCheckpointConnector,
)

log = logging.getLogger(__name__)


class PTLFlowTrainer(pl.Trainer):
    @_defaults_from_env_vars
    def __init__(
        self,
        *,
        accelerator: Union[str, Accelerator] = "auto",
        strategy: Union[str, Strategy] = "auto",
        devices: Union[List[int], str, int] = "auto",
        num_nodes: int = 1,
        precision: Optional[_PRECISION_INPUT] = None,
        logger: Optional[Union[Logger, Iterable[Logger], bool]] = None,
        callbacks: Optional[Union[List[Callback], Callback]] = None,
        fast_dev_run: Union[int, bool] = False,
        max_epochs: Optional[int] = None,
        min_epochs: Optional[int] = None,
        max_steps: int = -1,
        min_steps: Optional[int] = None,
        max_time: Optional[Union[str, timedelta, Dict[str, int]]] = None,
        limit_train_batches: Optional[Union[int, float]] = None,
        limit_val_batches: Optional[Union[int, float]] = None,
        limit_test_batches: Optional[Union[int, float]] = None,
        limit_predict_batches: Optional[Union[int, float]] = None,
        overfit_batches: Union[int, float] = 0.0,
        val_check_interval: Optional[Union[int, float]] = None,
        check_val_every_n_epoch: Optional[int] = 1,
        num_sanity_val_steps: Optional[int] = None,
        log_every_n_steps: Optional[int] = None,
        enable_checkpointing: Optional[bool] = None,
        enable_progress_bar: Optional[bool] = None,
        enable_model_summary: Optional[bool] = None,
        accumulate_grad_batches: int = 1,
        gradient_clip_val: Optional[Union[int, float]] = None,
        gradient_clip_algorithm: Optional[str] = None,
        deterministic: Optional[Union[bool, _LITERAL_WARN]] = None,
        benchmark: Optional[bool] = None,
        inference_mode: bool = True,
        use_distributed_sampler: bool = True,
        profiler: Optional[Union[Profiler, str]] = None,
        detect_anomaly: bool = False,
        barebones: bool = False,
        plugins: Optional[Union[_PLUGIN_INPUT, List[_PLUGIN_INPUT]]] = None,
        sync_batchnorm: bool = False,
        reload_dataloaders_every_n_epochs: int = 0,
        default_root_dir: Optional[_PATH] = None,
    ) -> None:
        r"""Customize every aspect of training via flags.

        Args:
            accelerator: Supports passing different accelerator types ("cpu", "gpu", "tpu", "hpu", "mps", "auto")
                as well as custom accelerator instances.

            strategy: Supports different training strategies with aliases as well custom strategies.
                Default: ``"auto"``.

            devices: The devices to use. Can be set to a positive number (int or str), a sequence of device indices
                (list or str), the value ``-1`` to indicate all available devices should be used, or ``"auto"`` for
                automatic selection based on the chosen accelerator. Default: ``"auto"``.

            num_nodes: Number of GPU nodes for distributed training.
                Default: ``1``.

            precision: Double precision (64, '64' or '64-true'), full precision (32, '32' or '32-true'),
                16bit mixed precision (16, '16', '16-mixed') or bfloat16 mixed precision ('bf16', 'bf16-mixed').
                Can be used on CPU, GPU, TPUs, or HPUs.
                Default: ``'32-true'``.

            logger: Logger (or iterable collection of loggers) for experiment tracking. A ``True`` value uses
                the default ``TensorBoardLogger`` if it is installed, otherwise ``CSVLogger``.
                ``False`` will disable logging. If multiple loggers are provided, local files
                (checkpoints, profiler traces, etc.) are saved in the ``log_dir`` of the first logger.
                Default: ``True``.

            callbacks: Add a callback or list of callbacks.
                Default: ``None``.

            fast_dev_run: Runs n if set to ``n`` (int) else 1 if set to ``True`` batch(es)
                of train, val and test to find any bugs (ie: a sort of unit test).
                Default: ``False``.

            max_epochs: Stop training once this number of epochs is reached. Disabled by default (None).
                If both max_epochs and max_steps are not specified, defaults to ``max_epochs = 1000``.
                To enable infinite training, set ``max_epochs = -1``.

            min_epochs: Force training for at least these many epochs. Disabled by default (None).

            max_steps: Stop training after this number of steps. Disabled by default (-1). If ``max_steps = -1``
                and ``max_epochs = None``, will default to ``max_epochs = 1000``. To enable infinite training, set
                ``max_epochs`` to ``-1``.

            min_steps: Force training for at least these number of steps. Disabled by default (``None``).

            max_time: Stop training after this amount of time has passed. Disabled by default (``None``).
                The time duration can be specified in the format DD:HH:MM:SS (days, hours, minutes seconds), as a
                :class:`datetime.timedelta`, or a dictionary with keys that will be passed to
                :class:`datetime.timedelta`.

            limit_train_batches: How much of training dataset to check (float = fraction, int = num_batches).
                Default: ``1.0``.

            limit_val_batches: How much of validation dataset to check (float = fraction, int = num_batches).
                Default: ``1.0``.

            limit_test_batches: How much of test dataset to check (float = fraction, int = num_batches).
                Default: ``1.0``.

            limit_predict_batches: How much of prediction dataset to check (float = fraction, int = num_batches).
                Default: ``1.0``.

            overfit_batches: Overfit a fraction of training/validation data (float) or a set number of batches (int).
                Default: ``0.0``.

            val_check_interval: How often to check the validation set. Pass a ``float`` in the range [0.0, 1.0] to check
                after a fraction of the training epoch. Pass an ``int`` to check after a fixed number of training
                batches. An ``int`` value can only be higher than the number of training batches when
                ``check_val_every_n_epoch=None``, which validates after every ``N`` training batches
                across epochs or during iteration-based training.
                Default: ``1.0``.

            check_val_every_n_epoch: Perform a validation loop after every `N` training epochs. If ``None``,
                validation will be done solely based on the number of training batches, requiring ``val_check_interval``
                to be an integer value.
                Default: ``1``.

            num_sanity_val_steps: Sanity check runs n validation batches before starting the training routine.
                Set it to `-1` to run all batches in all validation dataloaders.
                Default: ``2``.

            log_every_n_steps: How often to log within steps.
                Default: ``50``.

            enable_checkpointing: If ``True``, enable checkpointing.
                It will configure a default ModelCheckpoint callback if there is no user-defined ModelCheckpoint in
                :paramref:`~lightning.pytorch.trainer.trainer.Trainer.callbacks`.
                Default: ``True``.

            enable_progress_bar: Whether to enable to progress bar by default.
                Default: ``True``.

            enable_model_summary: Whether to enable model summarization by default.
                Default: ``True``.

            accumulate_grad_batches: Accumulates gradients over k batches before stepping the optimizer.
                Default: 1.

            gradient_clip_val: The value at which to clip gradients. Passing ``gradient_clip_val=None`` disables
                gradient clipping. If using Automatic Mixed Precision (AMP), the gradients will be unscaled before.
                Default: ``None``.

            gradient_clip_algorithm: The gradient clipping algorithm to use. Pass ``gradient_clip_algorithm="value"``
                to clip by value, and ``gradient_clip_algorithm="norm"`` to clip by norm. By default it will
                be set to ``"norm"``.

            deterministic: If ``True``, sets whether PyTorch operations must use deterministic algorithms.
                Set to ``"warn"`` to use deterministic algorithms whenever possible, throwing warnings on operations
                that don't support deterministic mode. If not set, defaults to ``False``. Default: ``None``.

            benchmark: The value (``True`` or ``False``) to set ``torch.backends.cudnn.benchmark`` to.
                The value for ``torch.backends.cudnn.benchmark`` set in the current session will be used
                (``False`` if not manually set). If :paramref:`~lightning.pytorch.trainer.trainer.Trainer.deterministic`
                is set to ``True``, this will default to ``False``. Override to manually set a different value.
                Default: ``None``.

            inference_mode: Whether to use :func:`torch.inference_mode` or :func:`torch.no_grad` during
                evaluation (``validate``/``test``/``predict``).

            use_distributed_sampler: Whether to wrap the DataLoader's sampler with
                :class:`torch.utils.data.DistributedSampler`. If not specified this is toggled automatically for
                strategies that require it. By default, it will add ``shuffle=True`` for the train sampler and
                ``shuffle=False`` for validation/test/predict samplers. If you want to disable this logic, you can pass
                ``False`` and add your own distributed sampler in the dataloader hooks. If ``True`` and a distributed
                sampler was already added, Lightning will not replace the existing one. For iterable-style datasets,
                we don't do this automatically.

            profiler: To profile individual steps during training and assist in identifying bottlenecks.
                Default: ``None``.

            detect_anomaly: Enable anomaly detection for the autograd engine.
                Default: ``False``.

            barebones: Whether to run in "barebones mode", where all features that may impact raw speed are
                disabled. This is meant for analyzing the Trainer overhead and is discouraged during regular training
                runs. The following features are deactivated:
                :paramref:`~lightning.pytorch.trainer.trainer.Trainer.enable_checkpointing`,
                :paramref:`~lightning.pytorch.trainer.trainer.Trainer.logger`,
                :paramref:`~lightning.pytorch.trainer.trainer.Trainer.enable_progress_bar`,
                :paramref:`~lightning.pytorch.trainer.trainer.Trainer.log_every_n_steps`,
                :paramref:`~lightning.pytorch.trainer.trainer.Trainer.enable_model_summary`,
                :paramref:`~lightning.pytorch.trainer.trainer.Trainer.num_sanity_val_steps`,
                :paramref:`~lightning.pytorch.trainer.trainer.Trainer.fast_dev_run`,
                :paramref:`~lightning.pytorch.trainer.trainer.Trainer.detect_anomaly`,
                :paramref:`~lightning.pytorch.trainer.trainer.Trainer.profiler`,
                :meth:`~lightning.pytorch.core.LightningModule.log`,
                :meth:`~lightning.pytorch.core.LightningModule.log_dict`.
            plugins: Plugins allow modification of core behavior like ddp and amp, and enable custom lightning plugins.
                Default: ``None``.

            sync_batchnorm: Synchronize batch norm layers between process groups/whole world.
                Default: ``False``.

            reload_dataloaders_every_n_epochs: Set to a positive integer to reload dataloaders every n epochs.
                Default: ``0``.

            default_root_dir: Default path for logs and weights when no logger/ckpt_callback passed.
                Default: ``os.getcwd()``.
                Can be remote file paths such as `s3://mybucket/path` or 'hdfs://path/'

        Raises:
            TypeError:
                If ``gradient_clip_val`` is not an int or float.

            MisconfigurationException:
                If ``gradient_clip_algorithm`` is invalid.

        """
        super().__init__(
            accelerator=accelerator,
            strategy=strategy,
            devices=devices,
            num_nodes=num_nodes,
            precision=precision,
            logger=logger,
            callbacks=callbacks,
            fast_dev_run=fast_dev_run,
            max_epochs=max_epochs,
            min_epochs=min_epochs,
            max_steps=max_steps,
            min_steps=min_steps,
            max_time=max_time,
            limit_train_batches=limit_train_batches,
            limit_val_batches=limit_val_batches,
            limit_test_batches=limit_test_batches,
            limit_predict_batches=limit_predict_batches,
            overfit_batches=overfit_batches,
            val_check_interval=val_check_interval,
            check_val_every_n_epoch=check_val_every_n_epoch,
            num_sanity_val_steps=num_sanity_val_steps,
            log_every_n_steps=log_every_n_steps,
            enable_checkpointing=enable_checkpointing,
            enable_progress_bar=enable_progress_bar,
            enable_model_summary=enable_model_summary,
            accumulate_grad_batches=accumulate_grad_batches,
            gradient_clip_val=gradient_clip_val,
            gradient_clip_algorithm=gradient_clip_algorithm,
            deterministic=deterministic,
            benchmark=benchmark,
            inference_mode=inference_mode,
            use_distributed_sampler=use_distributed_sampler,
            profiler=profiler,
            detect_anomaly=detect_anomaly,
            barebones=barebones,
            plugins=plugins,
            sync_batchnorm=sync_batchnorm,
            reload_dataloaders_every_n_epochs=reload_dataloaders_every_n_epochs,
            default_root_dir=default_root_dir,
        )
        self._checkpoint_connector = _PTLFlowCheckpointConnector(self)

    def _run(
        self, model: "pl.LightningModule", ckpt_path: Optional[_PATH] = None
    ) -> Optional[Union[_EVALUATE_OUTPUT, _PREDICT_OUTPUT]]:
        if self.state.fn == TrainerFn.FITTING:
            min_epochs, max_epochs = _parse_loop_limits(
                self.min_steps, self.max_steps, self.min_epochs, self.max_epochs, self
            )
            self.fit_loop.min_epochs = min_epochs
            self.fit_loop.max_epochs = max_epochs

        if self.barebones:
            # no progress bar in barebones can make it look like the Trainer hung
            rank_zero_info(
                "`Trainer(barebones=True)` started running. The progress bar is disabled so you might want to"
                " manually print the progress in your model."
            )

        # clean hparams
        if hasattr(model, "hparams"):
            parsing.clean_namespace(model.hparams)

        # attach model to the strategy
        self.strategy.connect(model)

        self._callback_connector._attach_model_callbacks()
        self._callback_connector._attach_model_logging_functions()

        _verify_loop_configurations(self)

        # ----------------------------
        # SET UP THE TRAINER
        # ----------------------------
        log.debug(f"{self.__class__.__name__}: setting up strategy environment")
        self.strategy.setup_environment()
        self.__setup_profiler()

        log.debug(f"{self.__class__.__name__}: preparing data")
        self._data_connector.prepare_data()

        call._call_setup_hook(
            self
        )  # allow user to set up LightningModule in accelerator environment
        log.debug(f"{self.__class__.__name__}: configuring model")
        call._call_configure_model(self)

        # check if we should delay restoring checkpoint till later
        if not self.strategy.restore_checkpoint_after_setup:
            log.debug(
                f"{self.__class__.__name__}: restoring module and callbacks from checkpoint path: {ckpt_path}"
            )
            self._checkpoint_connector._restore_modules_and_callbacks(ckpt_path, model)

        # reset logger connector
        self._logger_connector.reset_results()
        self._logger_connector.reset_metrics()

        # strategy will configure model and move it to the device
        self.strategy.setup(self)

        # hook
        if self.state.fn == TrainerFn.FITTING:
            call._call_callback_hooks(self, "on_fit_start")
            call._call_lightning_module_hook(self, "on_fit_start")

        _log_hyperparams(self)

        if self.strategy.restore_checkpoint_after_setup:
            log.debug(
                f"{self.__class__.__name__}: restoring module and callbacks from checkpoint path: {ckpt_path}"
            )
            self._checkpoint_connector._restore_modules_and_callbacks(ckpt_path)

        # restore optimizers, etc.
        try:
            log.debug(f"{self.__class__.__name__}: restoring training state")
            self._checkpoint_connector.restore_training_state()
        except KeyError:
            log.info(
                "The provided checkpoint does not contain the training state. Only the model weights will be loaded."
            )

        self._checkpoint_connector.resume_end()

        self._signal_connector.register_signal_handlers()

        # ----------------------------
        # RUN THE TRAINER
        # ----------------------------
        results = self._run_stage()

        # ----------------------------
        # POST-Training CLEAN UP
        # ----------------------------
        log.debug(f"{self.__class__.__name__}: trainer tearing down")
        self._teardown()

        if self.state.fn == TrainerFn.FITTING:
            call._call_callback_hooks(self, "on_fit_end")
            call._call_lightning_module_hook(self, "on_fit_end")

        log.debug(f"{self.__class__.__name__}: calling teardown hooks")
        call._call_teardown_hook(self)

        self.state.status = TrainerStatus.FINISHED
        self.state.stage = None

        return results

    def __setup_profiler(self) -> None:
        assert self.state.fn is not None
        local_rank = self.local_rank if self.world_size > 1 else None
        self.profiler._lightning_module = proxy(self.lightning_module)
        self.profiler.setup(
            stage=self.state.fn, local_rank=local_rank, log_dir=self.log_dir
        )
